"""
Absfuyu: Beautiful
------------------
A decorator that makes output more beautiful

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "BeautifulOutput",
    "print",
]


# Library
# ---------------------------------------------------------------------------
import time
import tracemalloc
from collections.abc import Callable
from functools import wraps
from typing import Any, Literal, NamedTuple, ParamSpec, TypeVar

BEAUTIFUL_MODE = False

try:
    from rich.align import Align
    from rich.console import Console, Group
    from rich.panel import Panel
    from rich.table import Table
    from rich.text import Text
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:
        cmd = "python -m pip install -U absfuyu[beautiful]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[beautiful] package")  # noqa: B904
else:
    BEAUTIFUL_MODE = True

# Setup
# ---------------------------------------------------------------------------
# rich's console.print wrapper
console = Console(color_system="auto", tab_size=4)
print = console.print


# Type
# ---------------------------------------------------------------------------
P = ParamSpec("P")  # Parameter type
R = TypeVar("R")  # Return type - Can be anything
T = TypeVar("T", bound=type)  # Type type - Can be any subtype of `type`


# Class
# ---------------------------------------------------------------------------
class PerformanceOutput(NamedTuple):
    runtime: float
    current_memory: int
    peak_memory: int

    def to_text(self) -> str:
        """
        Beautify the result and ready to print
        """
        out = (
            f"Memory usage:      {self.current_memory / 10**6:,.6f} MB\n"
            f"Peak memory usage: {self.peak_memory / 10**6:,.6f} MB\n"
            f"Time elapsed:      {self.runtime:,.6f} s"
        )
        return out


# TODO: header and footer layout to 1,2,3 instead of true false
class BeautifulOutput:
    """A decorator that makes output more beautiful"""

    def __init__(
        self,
        layout: Literal[1, 2, 3, 4, 5, 6] = 1,
        include_header: bool = True,
        include_footer: bool = True,
        alternate_footer: bool = False,
    ) -> None:
        """
        Show function's signature and measure memory usage

        Parameters
        ----------
        layout : Literal[1, 2, 3, 4, 5, 6], optional
            Layout to show, by default ``1``

        include_header : bool, optional
            Include header with function's signature, by default ``True``

        include_footer : bool, optional
            Include footer, by default ``True``

        alternate_footer : bool, optional
            Alternative style of footer, by default ``False``


        Usage
        -----
        Use this as a decorator (``@BeautifulOutput(<parameters>)``)
        """
        self.layout = layout
        self.include_header = include_header
        self.include_footer = include_footer
        self.alternate_footer = alternate_footer

        # Data
        self._obj_name = ""
        self._signature = ""
        self._result: Any | None = None
        self._performance: PerformanceOutput | None = None

        # Setting
        self._header_footer_style = "white on blue"
        self._alignment = "center"

    def __call__(self, obj: Callable[P, R]) -> Callable[P, Group]:
        # Class wrapper
        if isinstance(obj, type):
            raise NotImplementedError("Classes are not supported")

        # Function wrapper
        @wraps(obj)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Group:
            """
            Wrapper function that executes the original function.
            """
            # Get all parameters inputed
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={repr(v)}" for k, v in kwargs.items()]
            self._signature = ", ".join(args_repr + kwargs_repr)
            self._obj_name = obj.__name__

            # Performance check
            tracemalloc.start()  # Start memory measure
            start_time = time.perf_counter()  # Start time measure

            self._result = obj(*args, **kwargs)  # Function run

            finish_time = time.perf_counter()  # Get finished time
            _cur, _peak = tracemalloc.get_traced_memory()  # Get memory stats
            tracemalloc.stop()  # End memory measure

            self._performance = PerformanceOutput(
                runtime=finish_time - start_time, current_memory=_cur, peak_memory=_peak
            )

            return self._get_layout(layout=self.layout)

        return wrapper

    # Signature
    def _func_signature(self) -> str:
        """Function's signature"""
        return f"{self._obj_name}({self._signature})"

    # Layout
    def _make_header(self) -> Table:
        header_table = Table.grid(expand=True)
        header_table.add_row(
            Panel(
                Align(f"[b]{self._func_signature()}", align=self._alignment),
                style=self._header_footer_style,
            )
        )
        return header_table

    def _make_line(self) -> Table:
        line = Table.grid(expand=True)
        line.add_row(Text("", style=self._header_footer_style))
        return line

    def _make_footer(self) -> Table:
        if self.alternate_footer:
            return self._make_line()

        footer_table = Table.grid(expand=True)
        footer_table.add_row(
            Panel(
                Align("[b]BeautifulOutput by absfuyu", align=self._alignment),
                style=self._header_footer_style,
            )
        )
        return footer_table

    def _make_result_panel(self) -> Panel:
        result_txt = Text(
            str(self._result),
            overflow="fold",
            no_wrap=False,
            tab_size=2,
        )
        result_panel = Panel(
            Align(result_txt, align=self._alignment),
            title="[bold]Result[/]",
            border_style="green",
            highlight=True,
        )
        return result_panel

    def _make_performance_panel(self) -> Panel:
        if self._performance is not None:
            performance_panel = Panel(
                Align(self._performance.to_text(), align=self._alignment),
                title="[bold]Performance[/]",
                border_style="red",
                highlight=True,
                # height=result_panel.height,
            )
            return performance_panel
        else:
            return Panel("None", title="[bold]Performance[/]")

    def _make_output(self) -> Table:
        out_table = Table.grid(expand=True)
        out_table.add_column(ratio=3)  # result
        out_table.add_column(ratio=2)  # performance

        out_table.add_row(
            self._make_result_panel(),
            self._make_performance_panel(),
        )
        return out_table

    def _get_layout(self, layout: int) -> Group:
        header = self._make_header() if self.include_header else Text()
        footer = self._make_footer() if self.include_footer else Text()
        layouts = {
            1: Group(header, self._make_output(), footer),
            2: Group(header, self._make_result_panel(), self._make_performance_panel()),
            3: Group(header, self._make_result_panel(), footer),
            4: Group(self._make_result_panel(), self._make_performance_panel()),
            5: Group(self._make_output()),
            6: Group(
                header,
                self._make_result_panel(),
                self._make_performance_panel(),
                footer,
            ),
        }
        return layouts.get(layout, layouts[1])



"""
Absfuyu: Data Extension
-----------------------
Base type expansion

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# from __future__ import annotations

# Module Package
# ---------------------------------------------------------------------------
__all__ = []


# Library
# ---------------------------------------------------------------------------
from abc import ABC
from typing import Generic, Self, TypeVar

# Class
# ---------------------------------------------------------------------------
T = TypeVar("T")


class BaseType(Generic[T], ABC):
    """
    A universal base class for creating custom immutable type wrappers.

    Features:
    ----------
    - Works with any immutable type (int, str, float, etc.)
    - Provides validation hook via `_validate`
    - Automatically infers the base type from generic argument


    Example:
    --------
    >>> class NonNegativeInt(BaseType[int], int):
    >>>     @classmethod
    >>>     def _validate(cls, value: int) -> None:
    >>>         if value < 0:
    >>>             raise ValueError("Value must be non-negative")
    """

    # Automatically infer `_base_type` when subclassed
    def __init_subclass__(cls, *args, **kwargs) -> None:
        super().__init_subclass__(*args, **kwargs)

        # print(getattr(cls, "__orig_bases__", []))
        for base in getattr(cls, "__orig_bases__", []):
            if hasattr(base, "__args__") and base.__args__:
                cls._base_type: type[T] = base.__args__[0]
                break
        else:
            raise TypeError(f"{cls.__name__} must specify a generic base type, e.g. BaseType[int]")

    def __new__(cls, value: T) -> Self:
        base_type = cls._base_type
        if not isinstance(value, base_type):
            raise TypeError(f"{cls.__name__} must be initialized with {base_type.__name__}, got {type(value).__name__}")
        cls._validate(value)
        return base_type.__new__(cls, value)  # type: ignore

    @classmethod
    def _validate(cls, value: T) -> None:
        """Override this in subclasses to add custom validation."""
        pass

    def __repr__(self) -> str:
        # base_name = self._base_type.__name__
        # return f"{self.__class__.__name__}({base_name}={self._base_type(self)!r})"  # type: ignore
        return f"{self.__class__.__name__}({self._base_type(self)!r})"  # type: ignore


class IntBase(BaseType[int], int):
    """
    A base class for creating custom integer-like types.
    Provides a hook for validation and extension.

    Parameters
    ----------
    value : int
        Int value


    Usage:
    ------
    Use ``_validate(cls, value: int) -> None:`` classmethod to create custom validator
    """

    pass



"""
Test: Inspector

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

from typing import Any

import pytest

from absfuyu.core import BaseClass
from absfuyu.tools.inspector import Inspector
from absfuyu.typings import T as _T


def dummy_function(param: _T, *args, **kwargs):
    """This is a docstring"""
    pass


@pytest.fixture
def test_class():
    class TestClass(BaseClass):
        """a docs

        Extra part
        that very longgggggggggggggggggggggggggggggggggggggg
        gggggggggggggggggggggggggggggggggggggggggggggggggggg
        """

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            self.help = True
            self._hidden = True
            self.__hidden = True
            self.args = args

            for k, v in kwargs.items():
                setattr(self, k, v)

        @property
        def property_name(self): ...

        @property
        def _property_name(self): ...

        @property
        def __property_name(self): ...

        @classmethod
        def class_method_name(cls): ...

        @classmethod
        def _class_method_name(cls): ...

        @classmethod
        def __class_method_name(cls): ...

        @staticmethod
        def static_method_name(): ...

        @staticmethod
        def _static_method_name(): ...

        @staticmethod
        def __static_method_name(): ...

        def normal_method(self): ...
        def _normal_method(self): ...
        def __normal_method(self): ...

    return TestClass


@pytest.mark.abs_tools
class TestInspector:
    def test_1(self, test_class) -> None:
        pass



"""
Absfuyu: Data Extension
-----------------------
Extension for data type such as ``list``, ``str``, ``dict``, ...

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)

Features:
---------
- DictExt
- IntExt
- ListExt
- Text
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    # Main
    "DictExt",
    "IntExt",
    "ListExt",
    "Text",
    # Support
    "DictAnalyzeResult",
    "DictBoolFalse",
    "DictBoolTrue",
    "ListNoDunder",
    "ListREPR",
    "Pow",
    "TextAnalyzeDictFormat",
]


# Library
# ---------------------------------------------------------------------------
from absfuyu.dxt.dictext import DictAnalyzeResult, DictExt
from absfuyu.dxt.dxt_support import DictBoolFalse, DictBoolTrue, ListNoDunder, ListREPR
from absfuyu.dxt.intext import IntExt, Pow
from absfuyu.dxt.listext import ListExt
from absfuyu.dxt.strext import Text, TextAnalyzeDictFormat



"""
Absfuyu: Shorten number
-----------------------
Short number base on suffixes

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "UnitSuffixFactory",
    "CommonUnitSuffixesFactory",
    "Decimal",
    "shorten_number",
]


# Library
# ---------------------------------------------------------------------------
from collections.abc import Callable
from dataclasses import dataclass, field
from functools import wraps
from typing import Annotated, NamedTuple, ParamSpec, Self, TypeVar

from absfuyu.core import versionadded

# Type
# ---------------------------------------------------------------------------
P = ParamSpec("P")  # Parameter type
N = TypeVar("N", int, float)  # Number type


# Class
# ---------------------------------------------------------------------------
@versionadded("4.1.0")
class UnitSuffixFactory(NamedTuple):
    base: int
    short_name: list[str]
    full_name: list[str]


@versionadded("4.1.0")
class CommonUnitSuffixesFactory:
    NUMBER = UnitSuffixFactory(
        1000,
        [
            "",
            "K",
            "M",
            "B",
            "T",
            "Qa",
            "Qi",
            "Sx",
            "Sp",
            "Oc",
            "No",
            "Dc",
            "Ud",
            "Dd",
            "Td",
            "Qad",
            "Qid",
            "Sxd",
            "Spd",
            "Ocd",
            "Nod",
            "Vg",
            "Uvg",
            "Dvg",
            "Tvg",
            "Qavg",
            "Qivg",
            "Sxvg",
            "Spvg",
            "Ovg",
            "Nvg",
            "Tg",
            "Utg",
            "Dtg",
            "Ttg",
            "Qatg",
            "Qitg",
            "Sxtg",
            "Sptg",
            "Otg",
            "Ntg",
        ],
        [
            "",  # < Thousand
            "Thousand",
            "Million",
            "Billion",  # 1e9
            "Trillion",
            "Quadrillion",
            "Quintillion",
            "Sextillion",
            "Septillion",
            "Octillion",
            "Nonillion",
            "Decillion",  # 1e33
            "Undecillion",
            "Duodecillion",
            "Tredecillion",
            "Quattuordecillion",
            "Quindecillion",
            "Sexdecillion",
            "Septendecillion",
            "Octodecillion",
            "Novemdecillion",
            "Vigintillion",  #  1e63
            "Unvigintillion",
            "Duovigintillion",
            "Tresvigintillion",
            "Quattuorvigintillion",
            "Quinvigintillion",
            "Sesvigintillion",
            "Septemvigintillion",
            "Octovigintillion",
            "Novemvigintillion",
            "Trigintillion",  # 1e93
            "Untrigintillion",
            "Duotrigintillion",
            "Trestrigintillion",
            "Quattuortrigintillion",
            "Quintrigintillion",
            "Sestrigintillion",
            "Septentrigintillion",
            "Octotrigintillion",
            "Noventrigintillion",  #  1e120
        ],
    )
    DATA_SIZE = UnitSuffixFactory(
        1024,
        ["b", "Kb", "MB", "GB", "TB", "PB", "EB", "ZB", "YB", "BB"],
        [
            "byte",
            "kilobyte",
            "megabyte",
            "gigabyte",
            "terabyte",
            "petabyte",
            "exabyte",
            "zetabyte",
            "yottabyte",
            "brontobyte",
        ],
    )


@dataclass
@versionadded("4.1.0")
class Decimal:
    """
    Shorten large number

    Parameters
    ----------
    original_value : int | float
        Value to shorten

    base : int
        Short by base (must be > 0)

    suffixes : list[str]
        List of suffixes to use (ascending order)

    factory : UnitSuffixFactory | None
        ``UnitSuffixFactory`` to use
        (will overwrite ``base`` and ``suffixes``)

    suffix_full_name : bool
        Use suffix full name (available with ``UnitSuffixFactory``), by default ``False``

    Returns
    -------
    Decimal
        Decimal instance
    """

    original_value: int | float = field(repr=False)
    base: Annotated[int, "positive", "not_zero"] = field(repr=False, default=1000)
    suffixes: list[str] = field(repr=False, default_factory=list)
    factory: UnitSuffixFactory | None = field(repr=False, default=None)
    suffix_full_name: bool = field(repr=False, default=False)
    # Post init
    value: int | float = field(init=False)
    suffix: str = field(init=False)

    def __post_init__(self) -> None:
        self.base = max(1, self.base)  # Make sure that base >= 1
        self._get_factory()
        self.value, self.suffix = self._convert_decimal()

    def __str__(self) -> str:
        return self.to_text().strip()

    @classmethod
    def number(cls, value: int | float, suffix_full_name: bool = False) -> Self:
        """Decimal for normal large number"""
        return cls(
            value,
            factory=CommonUnitSuffixesFactory.NUMBER,
            suffix_full_name=suffix_full_name,
        )

    @classmethod
    def data_size(cls, value: int | float, suffix_full_name: bool = False) -> Self:
        """Decimal for data size"""
        return cls(
            value,
            factory=CommonUnitSuffixesFactory.DATA_SIZE,
            suffix_full_name=suffix_full_name,
        )

    @staticmethod
    def scientific_short(value: int | float) -> str:
        """Short number in scientific format"""
        return f"{value:.2e}"

    def _get_factory(self) -> None:
        if self.factory is not None:
            self.base = self.factory.base
            self.suffixes = (
                self.factory.full_name
                if self.suffix_full_name
                else self.factory.short_name
            )

    def _convert_decimal(self) -> tuple[float, str]:
        """Convert to smaller number"""
        suffix = self.suffixes[0] if len(self.suffixes) > 0 else ""
        unit = 1
        for i, suffix in enumerate(self.suffixes):
            unit = self.base**i
            if self.original_value < unit * self.base:
                break
        output = self.original_value / unit
        return output, suffix

    def to_text(
        self, decimal: int = 2, *, separator: str = " ", float_only: bool = True
    ) -> str:
        """
        Convert to string

        Parameters
        ----------
        decimal : int, optional
            Round up to which decimal, by default ``2``

        separator : str, optional
            Character between value and suffix, by default ``" "``

        float_only : bool, optional
            Returns value as <float> instead of <int> when ``decimal = 0``, by default ``True``

        Returns
        -------
        str
            Decimal string
        """
        val = self.value.__round__(decimal)
        formatted_value = f"{val:,}"
        if not float_only and decimal == 0:
            formatted_value = f"{int(val):,}"
        return f"{formatted_value}{separator}{self.suffix}"


# Decorator
# ---------------------------------------------------------------------------
@versionadded("5.0.0")
def shorten_number(f: Callable[P, N]) -> Callable[P, Decimal]:
    """
    Shorten the number value by name

    Parameters
    ----------
    f : Callable[P, N]
        Function that return ``int`` or ``float``

    Returns
    -------
    Callable[P, Decimal]
        Function that return ``Decimal``


    Usage
    -----
    Use this as a decorator (``@shorten_number``)

    Example:
    --------
    >>> import random
    >>> @shorten_number
    >>> def big_num() -> int:
    ...     random.randint(100000000, 10000000000)
    >>> big_num()
    4.20 B
    """

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Decimal:
        value = Decimal.number(f(*args, **kwargs))
        return value

    return wrapper



"""
Test: Util - Shorten number

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.util.shorten_number import CommonUnitSuffixesFactory, Decimal


@pytest.mark.abs_util
class TestUtilShortenNumber:
    """absfuyu.util.shorten_number"""

    @pytest.mark.parametrize(["value", "output"], [(1000, 1.0), (1000000, 1.0)])
    def test_number(self, value: int | float, output: float) -> None:
        ins = Decimal.number(value)
        assert ins.value == output

    def test_number2(self) -> None:
        fac = CommonUnitSuffixesFactory.NUMBER
        unit = 1
        for i, suffix in enumerate(fac.short_name):
            unit = fac.base**i
            assert Decimal.number(unit).suffix == suffix

    def test_data_size(self) -> None:
        fac = CommonUnitSuffixesFactory.DATA_SIZE
        unit = 1
        for i, suffix in enumerate(fac.short_name):
            unit = fac.base**i
            assert Decimal.data_size(unit).suffix == suffix



"""
Test: Tools - Obfuscator

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import base64

import pytest

from absfuyu.tools.obfuscator import Obfuscator, StrShifter


@pytest.fixture
def obfuscator_instance():
    code = "print('Hello World')"
    return Obfuscator(code=code)


@pytest.fixture
def str_shifter():
    """Fixture for creating a StrShifter instance."""
    return StrShifter("abcde", shift_by=2)


@pytest.mark.abs_tools
class TestObfuscator:
    """absfuyu.tools.obfuscator"""

    def test_convert_to_base64_decode(self) -> None:
        assert eval(Obfuscator._convert_to_base64_decode("rot_13")) == "rot_13"

    def test_obfuscate(self, obfuscator_instance: Obfuscator) -> None:
        assert obfuscator_instance.obfuscate()


@pytest.mark.abs_tools
class TestStrShifter:
    """absfuyu.tools.obfuscator.StrShifter"""

    def test_shift(self, str_shifter: StrShifter) -> None:
        """Test shifting characters."""
        result = str_shifter.shift()
        assert result == "deabc"

    def test_shift_with_different_shift(self, str_shifter: StrShifter) -> None:
        """Test shifting with a different shift value."""
        str_shifter.shift_by = 1
        result = str_shifter.shift()
        assert result == "eabcd"

    def test_shift_with_large_shift_value(self) -> None:
        """Test shifting with a shift value larger than string length."""
        shifter = StrShifter("abcde", shift_by=7)
        result = shifter.shift()
        assert result == "deabc"  # Same as shifting by 2

    def test_shift_with_zero_shift(self) -> None:
        """Test shifting with a zero shift value."""
        shifter = StrShifter("abcde", shift_by=0)
        result = shifter.shift()
        assert result == "abcde"  # No change

    def test_empty_string(self) -> None:
        """Test behavior with an empty string."""
        shifter = StrShifter("", shift_by=3)
        result = shifter.shift()
        assert result == ""  # Should remain empty

    def test_single_character(self) -> None:
        """Test behavior with a single character string."""
        shifter = StrShifter("a", shift_by=5)
        result = shifter.shift()
        assert result == "a"  # Should remain the same

    def test_special_characters(self) -> None:
        """Test shifting with special characters."""
        shifter = StrShifter("!@#$%", shift_by=3)
        result = shifter.shift()
        assert result == "$#%@!"  # Shift special characters

    def test_numeric_characters(self) -> None:
        """Test shifting with numeric characters."""
        shifter = StrShifter("12345", shift_by=2)
        result = shifter.shift()
        assert result == "45123"  # Shift numeric characters

    def test_invalid_type(self) -> None:
        """Test behavior with invalid input type."""
        with pytest.raises(TypeError, match="Value must be an instance of str"):
            StrShifter(123, shift_by=2)  # Should raise TypeError for non-string input



"""
Test: Core

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

from inspect import getdoc
from typing import Any

import pytest

from absfuyu import __version__
from absfuyu.core import BaseClass
from absfuyu.core.baseclass import (
    AutoREPRMixin,
    ClassMembers,
    ClassMembersResult,
    GetClassMembersMixin,
)
from absfuyu.core.decorator import dummy_decorator, dummy_decorator_with_args
from absfuyu.core.docstring import (
    _SPHINX_DOCS_TEMPLATE,
    SphinxDocstring,
    SphinxDocstringMode,
)


class ClassToTestDocs:
    clsattr = None

    def method(self, *args, **kwargs):
        """Normal method"""
        pass

    @classmethod
    def cmethod(cls, *args, **kwargs):
        """classmethod"""
        pass

    @staticmethod
    def stmethod(*args, **kwargs):
        """staticmethod"""
        pass

    @property
    def prop(self) -> None:
        pass


# MARK: fixture
@pytest.fixture
def clsmember_test() -> ClassMembers:
    return ClassMembers(["a"], ["b"], ["c"], ["d"], ["e"], ["f"])


@pytest.fixture
def clsmemres_test() -> ClassMembersResult:
    return ClassMembersResult(
        ABC=ClassMembers(["a"], ["b"], ["c"], ["d"]),
        DEF=ClassMembers(["e"], ["f"], ["g"], ["h"]),
    )


# MARK: core.baseclass
@pytest.mark.abs_core
class TestClassMembers:
    """
    ``absfuyu.core.baseclass.ClassMembers``
    """

    def test_repr(self, clsmember_test: ClassMembers) -> None:
        assert (
            repr(clsmember_test)
            == "ClassMembers(methods=['a'], classmethods=['b'], staticmethods=['c'], properties=['d'], attributes=['e'], classattributes=['f'])"
        )

    @pytest.mark.parametrize(
        ["value", "output"],
        [
            (([], [], [], [], [], []), True),
            ((["a"], [], [], [], [], []), False),
            (([], ["b"], [], [], [], []), False),
            (([], [], ["c"], [], [], []), False),
            (([], [], [], ["d"], [], []), False),
            (([], [], [], [], ["e"], []), False),
            (([], [], [], [], [], ["f"]), False),
        ],
    )
    def test_is_empty(self, value: tuple[list, list, list, list], output: bool) -> None:
        assert ClassMembers(*value).is_empty() == output

    @pytest.mark.parametrize(
        ["include_method", "include_classmethod", "include_staticmethod", "expected"],
        [
            (True, True, True, ["a", "b <cls>", "c <stc>"]),
            (False, True, True, ["b <cls>", "c <stc>"]),
            (True, False, True, ["a", "c <stc>"]),
            (True, True, False, ["a", "b <cls>"]),
        ],
    )
    def test_pack(
        self,
        clsmember_test: ClassMembers,
        include_method: bool,
        include_classmethod: bool,
        include_staticmethod: bool,
        expected,
    ) -> None:
        packed = clsmember_test.pack(
            include_method=include_method,
            include_classmethod=include_classmethod,
            include_staticmethod=include_staticmethod,
        )
        assert packed.methods == expected

    @pytest.mark.parametrize(
        ["include_attribute", "include_classattribute", "expected"],
        [
            (True, True, ["e", "f <cls>"]),
            (False, True, ["f <cls>"]),
            (True, False, ["e"]),
            (False, False, []),
        ],
    )
    def test_pack_2(
        self,
        clsmember_test: ClassMembers,
        include_attribute: bool,
        include_classattribute: bool,
        expected,
    ) -> None:
        packed = clsmember_test.pack(
            include_attribute=include_attribute,
            include_classattribute=include_classattribute,
        )
        assert packed.attributes == expected

    def test_pack_custom_indicators(self, clsmember_test: ClassMembers) -> None:
        packed = clsmember_test.pack(
            classmethod_indicator="(classmethod)",
            staticmethod_indicator="(staticmethod)",
        )
        assert packed == ClassMembers(
            ["a", "b (classmethod)", "c (staticmethod)"],
            [],
            [],
            ["d"],
            ["e", "f <cls>"],
        )


@pytest.mark.abs_core
class TestClassMembersResult:
    """
    ``absfuyu.core.baseclass.ClassMembersResult``
    """

    @pytest.mark.parametrize(
        ["value", "output"],
        [
            ("methods", ["a", "e"]),
            ("classmethods", ["b", "f"]),
            ("staticmethods", ["c", "g"]),
            ("properties", ["d", "h"]),
        ],
    )
    def test_merge_value(
        self,
        clsmemres_test: ClassMembersResult,
        value: str,
        output: list[str],
    ) -> None:
        assert clsmemres_test._merge_value(value) == output

    def test_flatten_value(self, clsmemres_test: ClassMembersResult) -> None:
        flattened = clsmemres_test.flatten_value()
        assert flattened == ClassMembers(["a", "e"], ["b", "f"], ["c", "g"], ["d", "h"])

    @pytest.mark.parametrize(
        ["include_method", "include_classmethod", "include_staticmethod", "expected"],
        [
            (True, True, True, ["a", "b <classmethod>", "c <staticmethod>"]),
            (False, True, True, ["b <classmethod>", "c <staticmethod>"]),
            (True, False, True, ["a", "c <staticmethod>"]),
            (True, True, False, ["a", "b <classmethod>"]),
        ],
    )
    def test_pack_value_parameterized(
        self,
        clsmemres_test: ClassMembersResult,
        include_method: bool,
        include_classmethod: bool,
        include_staticmethod: bool,
        expected: list[str],
    ) -> None:
        packed = clsmemres_test.pack_value(
            include_method=include_method,
            include_classmethod=include_classmethod,
            include_staticmethod=include_staticmethod,
        )
        assert packed["ABC"].methods == expected

    @pytest.mark.parametrize(
        ["value", "output"],
        [
            ("methods", {"ABC": ["a"], "DEF": ["e"]}),
            ("classmethods", {"ABC": ["b"], "DEF": ["f"]}),
            ("staticmethods", {"ABC": ["c"], "DEF": ["g"]}),
            ("properties", {"ABC": ["d"], "DEF": ["h"]}),
        ],
    )
    def test_prioritize_value(
        self,
        clsmemres_test: ClassMembersResult,
        value: str,
        output: dict[str, list[str]],
    ) -> None:
        prioritized_methods = clsmemres_test.prioritize_value(value)
        assert prioritized_methods == output


@pytest.mark.abs_core
class TestGetClassMembersMixin:
    """
    ``absfuyu.core.baseclass.GetClassMembersMixin``
    """

    def test_get_method_prop(self) -> None:
        class TestClass(ClassToTestDocs, GetClassMembersMixin): ...

        result = TestClass._get_members(dunder=False)
        assert isinstance(result, ClassMembersResult)
        m = result.flatten_value()
        assert "method" in m.methods
        assert "cmethod" in m.classmethods
        assert "stmethod" in m.staticmethods
        assert "prop" in m.properties
        assert "clsattr" in m.classattributes


@pytest.mark.abs_core
class TestAutoREPRMixin:
    """
    ``absfuyu.core.baseclass.AutoREPRMixin``
    """

    def test_class_no_slots(self) -> None:
        class ClassNoSlots(AutoREPRMixin):
            def __init__(self, a) -> None:
                self.a = a

        instance = ClassNoSlots(1)
        name = instance.__class__.__name__
        expected = f"{name}(a={instance.a!r})"
        assert repr(instance) == expected

    def test_class_with_slots(self) -> None:
        class ClassWithSlots(AutoREPRMixin):
            __slots__ = ("a",)

            def __init__(self, a) -> None:
                self.a = a

        instance = ClassWithSlots(1)
        name = instance.__class__.__name__
        expected = f"{name}(a={instance.a!r})"
        assert repr(instance) == expected


@pytest.mark.abs_core
class TestBaseClass:
    """
    ``absfuyu.core.BaseClass``
    """

    def test_BaseClass(self) -> None:
        _ = BaseClass.show_all_methods(print_result=True)


# MARK: core.decorator
@pytest.mark.abs_core
class TestCoreDecorator:
    def test_dummy_decorator(self) -> None:
        # Define a dummy function to decorate
        def add(a, b):
            return a + b

        # Apply the decorator
        decorated_add = dummy_decorator(add)

        # Test if the decorated function behaves as expected
        assert decorated_add(2, 3) == 5

    def test_dummy_decorator_class(self) -> None:
        # Define a class to decorate (which does nothing)
        class MyClass:
            pass

        # Apply the decorator to the class (should return unchanged)
        DecoratedClass = dummy_decorator(MyClass)

        assert DecoratedClass is MyClass

    def test_dummy_decorator_with_args(self) -> None:
        def multiply(a, b):
            return a * b

        decorator_instance = dummy_decorator_with_args("arg1", kwarg="value")

        decorated_multiply = decorator_instance(multiply)

        assert decorated_multiply(4, 5) == 20

    def test_dummy_decorator_with_args_class(self) -> None:
        class MyOtherClass:
            pass

        decorator_instance_for_class = dummy_decorator_with_args("arg1", kwarg="value")
        DecoratedOtherClass = decorator_instance_for_class(MyOtherClass)

        assert DecoratedOtherClass is MyOtherClass


# MARK: core.docstring
@pytest.mark.abs_core
class TestSphinxDocstring:
    """
    ``absfuyu.core.docstring.SphinxDocstring``
    """

    @pytest.mark.parametrize(
        ["reason", "mode"],
        [
            (None, SphinxDocstringMode.ADDED),
            (None, SphinxDocstringMode.CHANGED),
            (None, SphinxDocstringMode.DEPRECATED),
            ("test", SphinxDocstringMode.ADDED),
            ("test", SphinxDocstringMode.CHANGED),
            ("test", SphinxDocstringMode.DEPRECATED),
        ],
    )
    def test_SphinxDocstring_function(
        self, reason: str | None, mode: SphinxDocstringMode
    ) -> None:
        # Create a function with decorator
        @SphinxDocstring(__version__, reason=reason, mode=mode)
        def demo_function(parameter: Any) -> Any:
            return parameter

        @SphinxDocstring(__version__, reason=reason, mode=mode)
        def demo_function_2(parameter: Any) -> Any:
            """This already has docs"""
            return parameter

        # Get template
        _reason = f": {reason}" if reason else ""
        template = _SPHINX_DOCS_TEMPLATE.substitute(
            line_break="",
            mode=mode.value,
            version=__version__,
            reason=_reason,
        )

        for func in [demo_function, demo_function_2]:
            # Get docstring
            docs: str = getdoc(func)
            # Assert
            assert docs.endswith(template)

    @pytest.mark.parametrize(
        ["reason", "mode"],
        [
            (None, SphinxDocstringMode.ADDED),
            (None, SphinxDocstringMode.CHANGED),
            (None, SphinxDocstringMode.DEPRECATED),
            ("test", SphinxDocstringMode.ADDED),
            ("test", SphinxDocstringMode.CHANGED),
            ("test", SphinxDocstringMode.DEPRECATED),
        ],
    )
    def test_SphinxDocstring_class(
        self, reason: str | None, mode: SphinxDocstringMode
    ) -> None:
        @SphinxDocstring(__version__, reason=reason, mode=mode)
        class KlassNoDoc(ClassToTestDocs): ...

        @SphinxDocstring(__version__, reason=reason, mode=mode)
        class KlassWithDoc(ClassToTestDocs):
            """
            This is a doc
            """

            pass

        @SphinxDocstring(__version__, reason=reason, mode=mode)
        class SubClass(KlassNoDoc): ...

        test_list: list[ClassToTestDocs] = [KlassNoDoc, KlassWithDoc, SubClass]
        for k in test_list:
            docs: str = getdoc(k)  # Get docstring
            _reason = f": {reason}" if reason else ""
            template = _SPHINX_DOCS_TEMPLATE.substitute(
                line_break="",
                mode=mode.value,
                version=__version__,
                reason=_reason,
            )  # Retrive template str

            to_test = all(
                [
                    k.__doc__.endswith(template),
                    k().__doc__.endswith(template),
                    k.cmethod.__doc__ == ClassToTestDocs.cmethod.__doc__,
                    k().cmethod.__doc__ == ClassToTestDocs().cmethod.__doc__,
                    k.stmethod.__doc__ == ClassToTestDocs.stmethod.__doc__,
                    k().stmethod.__doc__ == ClassToTestDocs().stmethod.__doc__,
                    k.method.__doc__ == ClassToTestDocs.method.__doc__,
                    k().method.__doc__ == ClassToTestDocs().method.__doc__,
                ]
            )
            assert docs.endswith(template)
            assert to_test



"""
Absfuyu: Tax calculator
-----------------------
Tax calculator

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["PersonalIncomeTaxCalculator"]


# Library
# ---------------------------------------------------------------------------
from dataclasses import dataclass
from typing import Literal, TypedDict, overload

from absfuyu.core.baseclass import BaseClass, BaseDataclass

# Class
# ---------------------------------------------------------------------------
TaxLevel = list[tuple[float | int | None, float]]


@dataclass
class TaxLevelResult(BaseDataclass):
    """
    Result for a single tax level.

    Parameters
    ----------
    lower : float
        Lower bound

    upper : float | None
        Upper bound. ``None`` means +infinite

    rate : float
        Tax percentage in decimal (0.1 for 10%)

    amount_taxed : float
        Amount to calculate in current tax level

    tax : float
        Tax amount in current tax level
    """

    lower: float
    upper: float | None
    rate: float
    amount_taxed: float
    tax: float


class TaxCalculationResult(TypedDict):
    """
    Tax calculation result.

    Parameters
    ----------
    gross_income : float
        The input gross income for the calculation.

    deductions : float, optional
        Amount to subtract from gross income before tax.

    taxable_income : float
        Gross income minus deductions.

    per_tax_level : list[TaxLevelResult]
        Detailed breakdown of income taxed at each tax level.

    gross_tax : float
        Total tax before credits.

    tax_credits : float, optional
        Amount subtracted from the computed tax after calculation.

    net_tax : float
        Total tax after applying credits.

    effective_rate : float
        Ratio of net tax to gross income (net_tax / gross_income).

    marginal_rate : float
        Tax rate applied to the last amount of money of taxable income.
    """

    gross_income: float
    deductions: float
    taxable_income: float
    per_tax_level: list[TaxLevelResult]
    gross_tax: float
    tax_credits: float
    net_tax: float
    effective_rate: float
    marginal_rate: float


class TaxCalculationResultRaw(TypedDict):
    """
    Tax calculation result.

    Parameters
    ----------
    gross_income : float
        The input gross income for the calculation.

    deductions : float, optional
        Amount to subtract from gross income before tax.

    taxable_income : float
        Gross income minus deductions.

    per_tax_level : list[TaxLevelResult]
        Detailed breakdown of income taxed at each tax level.

    gross_tax : float
        Total tax before credits.

    tax_credits : float, optional
        Amount subtracted from the computed tax after calculation.

    net_tax : float
        Total tax after applying credits.

    effective_rate : float
        Ratio of net tax to gross income (net_tax / gross_income).

    marginal_rate : float
        Tax rate applied to the last amount of money of taxable income.
    """

    gross_income: float
    deductions: float
    taxable_income: float
    per_tax_level: list[dict[str, float | None]]
    gross_tax: float
    tax_credits: float
    net_tax: float
    effective_rate: float
    marginal_rate: float


class PersonalIncomeTaxCalculator(BaseClass):
    """
    Progressive personal income tax calculator.

    Parameters
    ----------
    tax_levels : list[tuple[float | int | None, float]], optional
        Ordered list of tax levels.
        Each tax level is represented as a tuple of (upper_bound, rate),
        where ``upper_bound`` is the cumulative upper limit of the tax level.
        Use ``None`` for the last tax level (no upper limit).
        Example: ``[(10000, 0.10), (50000, 0.2), (None, 0.3)]``.
        Set to ``None`` to have 0% tax rate, by default ``None``

    deductions : float, optional
        Amount to subtract from gross income before tax, by default 0.0

    tax_credits : float, optional
        Amount subtracted from the computed tax after calculation, by default 0.0

    Attributes
    ----------
    gross_income : float
        The input gross income for the calculation.

    taxable_income : float
        Gross income minus deductions.

    per_tax_level : list[TaxLevelResult]
        Detailed breakdown of income taxed at each tax level.

    gross_tax : float
        Total tax before credits.

    net_tax : float
        Total tax after applying credits.

    effective_rate : float
        Ratio of net tax to gross income.

    marginal_rate : float
        Tax rate applied to the last amount of money of taxable income.


    Example:
    --------
    >>> tax_levels = [(100, 0.05), (None, 0.1)]
    >>> cal = PersonalIncomeTaxCalculator(tax_levels)
    >>> cal.calculate(500)
    >>> cal.to_dict(raw=True)
    {...}

    >>> cal.interpret_result()
    ===== Tax information =====
    Taxable income: 500.00 (500.00 - 0.00)
    - Level 1: 0.00 - 100.00 @ 5.0%: 100.00 -> tax 5.00
    - Level 2: 100.00 - 500.00 @ 10.0%: 400.00 -> tax 40.00
    Net tax: 45.00 (45.00 - 0.00)
    Effective rate: 9.00%
    Marginal rate: 10.0%
    """

    def __init__(self, tax_levels: TaxLevel | None = None, deductions: float = 0.0, tax_credits: float = 0.0) -> None:
        """
        Progressive personal income tax calculator.

        Parameters
        ----------
        tax_levels : list[tuple[float | int | None, float]], optional
            Ordered list of tax levels.
            Each tax level is represented as a tuple of (upper_bound, rate),
            where ``upper_bound`` is the cumulative upper limit of the tax level.
            Use ``None`` for the last tax level (no upper limit).
            Example: ``[(10000, 0.10), (50000, 0.2), (None, 0.3)]``.
            Set to ``None`` to have 0% tax rate, by default ``None``

        deductions : float, optional
            Amount to subtract from gross income before tax, by default 0.0

        tax_credits : float, optional
            Amount subtracted from the computed tax after calculation, by default 0.0
        """
        self.tax_levels = [] if tax_levels is None else tax_levels
        self.deductions = deductions
        self.tax_credits = tax_credits

        # Results populated after calculation
        self.gross_income: float = 0.0
        self.taxable_income: float = 0.0
        self.per_tax_level: list[TaxLevelResult] = []
        self.gross_tax: float = 0.0
        self.net_tax: float = 0.0
        self.effective_rate: float = 0.0
        self.marginal_rate: float = 0.0

    def calculate(self, gross_income: float) -> None:
        """
        Compute tax for a given gross income.

        Parameters
        ----------
        gross_income : float
            Total gross income (unless tax levels are defined otherwise).
        """
        if gross_income < 0:
            raise ValueError("gross_income must be non-negative")

        self.gross_income = gross_income
        self.taxable_income = max(0.0, gross_income - max(0.0, self.deductions))

        self.per_tax_level = []
        prev_upper = 0.0
        remaining = self.taxable_income
        self.gross_tax = 0.0
        self.marginal_rate = 0.0

        for upper, rate in self.tax_levels:
            lower = prev_upper
            if upper is None:
                amount_taxed = remaining
            else:
                band = upper - prev_upper
                amount_taxed = min(band, max(0.0, remaining))
            tax = max(0.0, amount_taxed) * rate
            self.per_tax_level.append(TaxLevelResult(lower, upper, rate, amount_taxed, tax))
            self.gross_tax += tax
            remaining -= amount_taxed
            prev_upper = upper if upper is not None else prev_upper
            if remaining <= 1e-9:
                self.marginal_rate = rate
                break
            self.marginal_rate = rate

        self.net_tax = max(0.0, self.gross_tax - max(0.0, self.tax_credits))
        self.effective_rate = self.net_tax / gross_income if gross_income > 0 else 0.0

    @overload
    def to_dict(self) -> TaxCalculationResult: ...  # type: ignore

    @overload
    def to_dict(self, *, raw: Literal[True] = ...) -> TaxCalculationResultRaw: ...

    def to_dict(self, *, raw: bool = False) -> TaxCalculationResult | TaxCalculationResultRaw:
        """
        Returns calculation result in dict format

        Parameters
        ----------
        raw : bool, optional
            Convert every value to dict, by default ``False``

        Returns
        -------
        TaxCalculationResult
            Tax calculation result
        """
        result: TaxCalculationResult = {
            "gross_income": self.gross_income,
            "deductions": self.deductions,
            "taxable_income": self.taxable_income,
            "per_tax_level": self.per_tax_level,
            "gross_tax": self.gross_tax,
            "tax_credits": self.tax_credits,
            "net_tax": self.net_tax,
            "effective_rate": self.effective_rate,
            "marginal_rate": self.marginal_rate,
        }
        if raw:
            result_raw: TaxCalculationResultRaw = result  # type: ignore
            result_raw["per_tax_level"] = [x.to_dict() for x in result["per_tax_level"]]
            return result
        return result

    def interpret_result(self) -> str:
        result = self.to_dict()
        text = ["===== Tax information ====="]

        # text.append(f"Gross income: {result['gross_income']:,}")
        # text.append(f"Deduction: {result['deductions']:,}")
        # text.append(f"Taxable income: {result['taxable_income']:,}")
        text.append(
            f"Taxable income: {result['taxable_income']:,.2f} ({result['gross_income']:,.2f} - {result['deductions']:,.2f})"
        )

        for idx, tax_level in enumerate(result["per_tax_level"], start=1):
            upper = f"{tax_level.upper:,.2f}" if tax_level.upper is not None else f"{self.gross_income:,.2f}"
            text.append(
                f"- Level {idx}: {tax_level.lower:,.2f} - {upper} @ {tax_level.rate*100:.1f}%: {tax_level.amount_taxed:,.2f} -> tax {tax_level.tax:,.2f}"
            )

        # text.append(f"Gross tax: {result['gross_tax']:,}")
        # text.append(f"Tax credits: {result['tax_credits']:,}")
        # text.append(f"Net tax: {result['net_tax']:,}")
        text.append(f"Net tax: {result['net_tax']:,.2f} ({result['gross_tax']:,.2f} - {result['tax_credits']:,.2f})")

        text.append(f"Effective rate: {result['effective_rate']*100:.2f}%")
        text.append(f"Marginal rate: {result['marginal_rate']*100:.1f}%")
        return "\n".join(text)



"""
ABSFUYU CLI
-----------
Tool

Version: 5.15.0
Date updated: 05/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["tool_group"]


# Library
# ---------------------------------------------------------------------------
from pathlib import Path
from typing import Literal

import click

from absfuyu.tools.checksum import Checksum
from absfuyu.tools.converter import Base64EncodeDecode, Text2Chemistry


# CLI
# ---------------------------------------------------------------------------
@click.command(name="cs")
@click.argument("file_path", type=str)
@click.option(
    "--hashmode",
    "-m",
    "hash_mode",
    type=click.Choice(["md5", "sha1", "sha256", "sha512"]),
    default="sha256",
    show_default=True,
    help="Hash mode",
)
@click.option(
    "--save-result",
    "-s",
    "save_result",
    type=bool,
    default=False,
    is_flag=True,
    show_default=True,
    help="Save checksum result to file",
)
@click.option(
    "--recursive",
    "-r",
    "recursive_mode",
    type=bool,
    default=False,
    is_flag=True,
    show_default=True,
    help="Do checksum for every file in the folder (including child folder)",
)
@click.option(
    "--compare",
    "-c",
    "hash_to_compare",
    type=str,
    default=None,
    show_default=True,
    help="Hash to compare",
)
def file_checksum(
    file_path: str,
    hash_mode: Literal["md5", "sha1", "sha256", "sha512"],
    save_result: bool,
    recursive_mode: bool,
    hash_to_compare: str | None,
) -> None:
    """Checksum for file/directory"""
    instance = Checksum(file_path, hash_mode=hash_mode, save_result_to_file=save_result)
    res = instance.checksum(recursive=recursive_mode)
    if hash_to_compare is None:
        click.echo(res)
    else:
        click.echo(res == hash_to_compare)


@click.command(name="cimg")
@click.argument("dir_path", type=str)
@click.argument("to_format", type=str)
def convert_img(dir_path: str, to_format: str) -> None:
    """Convert image in directory"""
    from absfuyu.extra.img.converter import ImgConverter

    engine = ImgConverter(Path(dir_path).resolve())
    engine.img_convert(to_format)
    click.echo("Done")


@click.command(name="rdimg")
@click.argument("dir_path", type=str)
@click.option(
    "--dryrun",
    "-d",
    "dry_run",
    type=bool,
    default=False,
    is_flag=True,
    show_default=True,
    help="Simulate deleting only",
)
@click.option(
    "--recursive",
    "-r",
    "recursive_mode",
    type=bool,
    default=False,
    is_flag=True,
    show_default=True,
    help="Scan for every file in the folder (including child folder)",
)
@click.option(
    "--threshold",
    "-t",
    "threshold",
    type=int,
    default=5,
    show_default=True,
    help="Threshold: 0 for exact image, 5 to 10 for light edit",
)
@click.option(
    "--keepmode",
    "-k",
    "keep_mode",
    type=click.Choice(["first", "last", "best"]),
    default="best",
    show_default=True,
    help="Keep mode",
)
def remove_dup_img(dir_path: str, threshold: bool, recursive_mode: bool, keep_mode: str, dry_run: bool) -> None:
    """Remove duplicate images in directory"""
    from absfuyu.extra.img.dup_check import DirectoryRemoveDuplicateImageMixin as DRDI

    engine = DRDI(Path(dir_path).resolve())
    engine.remove_duplicate_images(threshold=threshold, recursive=recursive_mode, keep_mode=keep_mode, dry_run=dry_run)
    click.echo("Done")


@click.command(name="t2c")
@click.argument("text", type=str)
def text2chem(text: str) -> None:
    """Convert text into chemistry symbol"""
    engine = Text2Chemistry()
    out = engine.convert(text)
    click.echo(Text2Chemistry.beautify_result(out))


@click.command(name="e")
@click.argument("text", type=str)
def base64encode(text: str) -> None:
    """Convert text to base64"""
    click.echo(Base64EncodeDecode.encode(text))


@click.command(name="d")
@click.argument("text", type=str)
def base64decode(text: str) -> None:
    """Convert base64 to text"""
    click.echo(Base64EncodeDecode.decode(text))


@click.command(name="img")
@click.option(
    "--data-tag",
    "-d",
    "data_tag",
    type=bool,
    default=False,
    is_flag=True,
    show_default=True,
    help="Add data tag before base64 string",
)
@click.argument("img_path", type=str)
def base64convert_img(img_path: str, data_tag: bool) -> None:
    """Convert img to base64"""
    click.echo(Base64EncodeDecode.encode_image(img_path, data_tag=data_tag))


@click.group(name="b64")
def base64_group() -> None:
    """Base64 encode decode"""
    pass


base64_group.add_command(base64encode)
base64_group.add_command(base64decode)
base64_group.add_command(base64convert_img)


@click.group(name="tool")
def tool_group() -> None:
    """Tools"""
    pass


tool_group.add_command(file_checksum)
tool_group.add_command(convert_img)
tool_group.add_command(remove_dup_img)
tool_group.add_command(base64_group)
tool_group.add_command(text2chem)



"""
Absfuyu: Data Extension
-----------------------
str extension

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["Text", "TextAnalyzeDictFormat"]


# Library
# ---------------------------------------------------------------------------
import random
from string import ascii_letters as _ascii_letters
from typing import NotRequired, Self, TypedDict

from absfuyu.core import GetClassMembersMixin, deprecated, versionadded, versionchanged
from absfuyu.util import set_min_max


# Support function
# ---------------------------------------------------------------------------
def _generate_string(
    charset: str | None = None,
    size: int = 8,
    times: int = 1,
    unique: bool = True,
) -> list[str]:
    """
    Generate a list of random string from character set
    (Random string generator)

    This is a lesser version of
    ``absfuyu.tools.generator.Generator.generate_string()``

    Parameters
    ----------
    charset : str, optional
        Custom character set, by default ``None``
        ([a-zA-Z] - string.ascii_letters)

    size : int, optional
            Length of each string in list, by default ``8``

    times : int, optional
        How many random string generated, by default ``1``

    unique : bool, optional
        Each generated text is unique, by default ``True``

    Returns
    -------
    list[str]
        List of random string generated
    """

    charset = _ascii_letters or charset

    try:
        char_lst = list(charset)  # type: ignore[arg-type]
    except Exception:
        char_lst = charset  # type: ignore[assignment]

    unique_string = []
    count = 0

    while count < times:
        gen_string = "".join(random.choice(char_lst) for _ in range(size))  # type: ignore
        if not unique:
            unique_string.append(gen_string)
            count += 1
        else:
            if gen_string not in unique_string:
                unique_string.append(gen_string)
                count += 1

    return unique_string


# Class
# ---------------------------------------------------------------------------
class TextAnalyzeDictFormat(TypedDict):
    """
    Dict format for ``Text.analyze()`` method

    Parameters
    ----------
    digit : int
        Number of digit characters

    uppercase : int
        Number of uppercase characters

    lowercase : int
        Number of lowercase characters

    other : int
        Number of other printable characters

    is_pangram : NotRequired[bool]
        Is a pangram (Not required)

    is_palindrome : NotRequired[bool]
        Is a palindrome (Not required)
    """

    digit: int
    uppercase: int
    lowercase: int
    other: int
    is_pangram: NotRequired[bool]
    is_palindrome: NotRequired[bool]


class Text(GetClassMembersMixin, str):
    """
    ``str`` extension

    >>> # For a list of new methods
    >>> Text.show_all_methods()
    """

    def divide(self, string_split_size: int = 60) -> list[str]:
        """
        Divide long string into smaller size

        Parameters
        ----------
        string_split_size : int
            Divide string every ``x`` character
            (Default: ``60``)

        Returns
        -------
        list[str]
            A list in which each item is a smaller
            string with the size of ``string_split_size``
            (need to be concaternate later)


        Example:
        --------
        >>> test = Text("This is an extremely long line of text!")
        >>> test.divide(string_split_size=20)
        ['This is an extremely', ' long line of text!']
        """
        temp = str(self)
        output = []
        while len(temp) != 0:
            output.append(temp[:string_split_size])
            temp = temp[string_split_size:]
        return output

    def divide_with_variable(
        self,
        split_size: int = 60,
        split_var_len: int = 12,
        custom_var_name: str | None = None,
    ) -> list[str]:
        """
        Divide long string into smaller size,
        then assign a random variable to splited
        string for later use

        Parameters
        ----------
        split_size : int
            Divide string every ``x`` character
            (Default: ``60``)

        split_var_len : int
            Length of variable name assigned to each item
            (Default: ``12``)

        custom_var_name : str
            Custom variable name when join string

        Returns
        -------
        list[str]
            A list in which each item is a smaller
            string with the size of ``split_size``
            and a way to concaternate them (when using ``print()``)


        Example:
        --------
        >>> test = Text("This is an extremely long line of text!")
        >>> test.divide_with_variable(split_size=20)
        [
            "qNTCnmkFPTJg='This is an extremely'",
            "vkmLBUykYYDG=' long line of text!'",
            'sBoSwEfoxBIH=qNTCnmkFPTJg+vkmLBUykYYDG',
            'sBoSwEfoxBIH'
        ]

        >>> test = Text("This is an extremely long line of text!")
        >>> test.divide_with_variable(split_size=20, custom_var_name="test")
        [
            "test1='This is an extremely'",
            "test2=' long line of text!'",
            'test=test1+test2',
            'test'
        ]
        """

        temp = self.divide(split_size)
        output = []

        # split variable
        splt_len = len(temp)

        if custom_var_name is None:
            splt_name = _generate_string(size=split_var_len, times=splt_len + 1)
            for i in range(splt_len):
                output.append(f"{splt_name[i]}='{temp[i]}'")
        else:
            for i in range(splt_len):
                output.append(f"{custom_var_name}{i + 1}='{temp[i]}'")

        # joined variable
        temp = []
        if custom_var_name is None:
            for i in range(splt_len):
                if i == 0:
                    temp.append(f"{splt_name[-1]}=")
                if i == splt_len - 1:
                    temp.append(f"{splt_name[i]}")
                else:
                    temp.append(f"{splt_name[i]}+")
        else:
            for i in range(splt_len):
                if i == 0:
                    temp.append(f"{custom_var_name}=")
                if i == splt_len - 1:
                    temp.append(f"{custom_var_name}{i + 1}")
                else:
                    temp.append(f"{custom_var_name}{i + 1}+")

        output.append("".join(temp))
        if custom_var_name is None:
            output.append(splt_name[-1])
        else:
            output.append(custom_var_name)

        return output

    @versionchanged("3.3.0", reason="Updated functionality")
    def analyze(self, full: bool = False) -> TextAnalyzeDictFormat:
        """
        String analyze (count number of type of character)

        Parameters
        ----------
        full : bool
            Full analyze when ``True``
            (Default: ``False``)

        Returns
        -------
        dict | TextAnalyzeDictFormat
            A dictionary contains number of digit character,
            uppercase character, lowercase character, and
            special character


        Example:
        --------
        >>> test = Text("Random T3xt!")
        >>> test.analyze()
        {'digit': 1, 'uppercase': 2, 'lowercase': 7, 'other': 2}
        """

        temp = self

        detail: TextAnalyzeDictFormat = {
            "digit": 0,
            "uppercase": 0,
            "lowercase": 0,
            "other": 0,
        }

        for x in temp:
            if ord(x) in range(48, 58):  # num
                detail["digit"] += 1
            elif ord(x) in range(65, 91):  # cap
                detail["uppercase"] += 1
            elif ord(x) in range(97, 123):  # low
                detail["lowercase"] += 1
            else:
                detail["other"] += 1

        if full:
            detail["is_palindrome"] = self.is_palindrome()
            detail["is_pangram"] = self.is_pangram()

        return detail

    def reverse(self) -> Self:
        """
        Reverse the string

        Returns
        -------
        Text
            Reversed string


        Example:
        --------
        >>> test = Text("Hello, World!")
        >>> test.reverse()
        '!dlroW ,olleH'
        """
        return self.__class__(self[::-1])

    @versionchanged("5.0.0", reason="Add ``custom_alphabet`` parameter")
    def is_pangram(self, custom_alphabet: set[str] | None = None) -> bool:
        """
        Check if string is a pangram

        A pangram is a unique sentence in which
        every letter of the alphabet is used at least once

        Parameters
        ----------
        custom_alphabet : set[str] | None, optional
            Custom alphabet to use
            (Default: ``None``)

        Returns
        -------
        bool
            | ``True`` if string is a pangram
            | ``False`` if string is not a pangram
        """
        text = self
        if custom_alphabet is None:
            alphabet = set("abcdefghijklmnopqrstuvwxyz")
        else:
            alphabet = custom_alphabet
        return not set(alphabet) - set(text.lower())

    def is_palindrome(self) -> bool:
        """
        Check if string is a palindrome

            A palindrome is a word, verse, or sentence
            or a number that reads the same backward or forward

        Returns
        -------
        bool
            | ``True`` if string is a palindrome
            | ``False`` if string is not a palindrome
        """
        text = self
        # Use string slicing [start:end:step]
        return text == text[::-1]

    def to_hex(self, raw: bool = False) -> str:
        r"""
        Convert string to hex form

        Parameters
        ----------
        raw : bool
            | ``False``: hex string in the form of ``\x`` (default)
            | ``True``: normal hex string

        Returns
        -------
        str
            Hexed string


        Example:
        --------
        >>> test = Text("Hello, World!")
        >>> test.to_hex()
        '\\x48\\x65\\x6c\\x6c\\x6f\\x2c\\x20\\x57\\x6f\\x72\\x6c\\x64\\x21'
        """
        text = self

        byte_str = text.encode("utf-8")
        # hex_str = byte_str.hex()

        if raw:
            return byte_str.hex()

        temp = byte_str.hex("x")
        return "\\x" + temp.replace("x", "\\x")

    def random_capslock(self, probability: int = 50) -> Self:
        """
        Randomly capslock letter in string

        Parameters
        ----------
        probability : int
            Probability in range [0, 100]
            (Default: ``50``)

        Returns
        -------
        Text
            Random capslocked text


        Example:
        --------
        >>> test = Text("This is an extremely long line of text!")
        >>> test.random_capslock()
        'tHis iS An ExtREmELY loNg liNE oF tExT!'
        """
        probability = int(set_min_max(probability))
        text = self.lower()

        random_caps = (
            x.upper() if random.randint(1, 100) <= probability else x for x in text
        )
        return self.__class__("".join(random_caps))

    @deprecated(
        "5.2.0",
        reason="str already has swapcase() method, will be removed in version 5.3.0",
    )
    @versionchanged("5.0.0", reason="Use ``str.swapcase()``")
    def reverse_capslock(self) -> Self:
        """
        Reverse capslock in string

        Returns
        -------
        Text
            Reversed capslock ``Text``


        Example:
        --------
        >>> test = Text("Foo")
        >>> test.reverse_capslock()
        'fOO'
        """
        return self.__class__(self.swapcase())

    def to_list(self) -> list[str]:
        """
        Convert into list

        Returns
        -------
        list[str]
            List of string


        Example:
        --------
        >>> test = Text("test")
        >>> test.to_list()
        ['t', 'e', 's', 't']
        """
        return list(self)

    @deprecated("5.0.0", reason="Unused, will be removed in version 5.3.0")
    def to_listext(self) -> None:
        """Deprecated, will be removed soon"""
        raise NotImplementedError("Deprecated, will be removed soon")

    @versionadded("3.3.0")
    def count_pattern(self, pattern: str, ignore_capslock: bool = False) -> int:
        """
        Returns how many times ``pattern`` appears in text

        Parameters
        ----------
        pattern : str
            Pattern to count

        ignore_capslock : bool
            Ignore the pattern uppercase or lowercase
            (Default: ``False`` - Exact match)

        Returns
        -------
        int
            How many times pattern appeared


        Example:
        --------
        >>> Text("test").count_pattern("t")
        2
        """
        if len(pattern) > len(self):
            raise ValueError(f"len(<pattern>) must not larger than {len(self)}")

        temp = str(self)
        if ignore_capslock:
            pattern = pattern.lower()
            temp = temp.lower()

        out = [
            1
            for i in range(len(temp) - len(pattern) + 1)
            if temp[i : i + len(pattern)] == pattern
        ]
        return sum(out)

    @versionadded("3.3.0")
    def hapax(self, strict: bool = False) -> list[str]:
        """
        A hapax legomenon (often abbreviated to hapax)
        is a word which occurs only once in either
        the written record of a language, the works of
        an author, or in a single text.

        This function returns a list of hapaxes (if any)
        (Lettercase is ignored)

        Parameters
        ----------
        strict : bool
            Remove all special characters before checking for hapax
            (Default: ``False``)

        Returns
        -------
        list[str]
            A list of hapaxes


        Example:
        --------
        >>> test = Text("A a. a, b c c= C| d d")
        >>> test.hapax()
        ['a', 'a.', 'a,', 'b', 'c', 'c=', 'c|']

        >>> test.hapax(strict=True)
        ['b']
        """
        word_list: list[str] = self.lower().split()
        if strict:
            remove_characters: list[str] = list(r"\"'.,:;|()[]{}\/!@#$%^&*-_=+?<>`~")
            temp = str(self)
            for x in remove_characters:
                temp = temp.replace(x, "")
            word_list = temp.lower().split()

        hapaxes = filter(lambda x: word_list.count(x) == 1, word_list)
        return list(hapaxes)

    @versionadded("5.0.0")
    def shorten(self, shorten_size: int = 60) -> str:
        """
        Shorten long text

        Parameters
        ----------
        shorten_size : int, optional
            How many characters per line.
            Minimum is ``1``, by default ``60``

        Returns
        -------
        str
            Shortened text


        Example:
        --------
        >>> test = Text("a" * 200)
        >>> test.shorten()
        (
        'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        'aaaaaaaaaaaaaaaaaaaa'
        )
        """
        shorten_text_list: list[str] = self.divide(
            string_split_size=max(1, shorten_size)
        )
        shorten_text_list = [repr(x) for x in shorten_text_list]
        out = "(\n" + "\n".join(shorten_text_list) + "\n)"
        return out



"""
Test: Data Analysis

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import random

import pytest

try:  # [extra] feature
    import numpy as np
    import pandas as pd
except ImportError:
    np = pytest.importorskip("numpy")
    pd = pytest.importorskip("pandas")

from absfuyu.extra.da.dadf import DADF
from absfuyu.extra.da.dadf_base import CityData, SplittedDF
from absfuyu.tools.generator import Charset, Generator

SAMPLE_SIZE = 100
sample_city_data = CityData._sample_city_data(size=SAMPLE_SIZE)


# MARK: fixture
@pytest.fixture
def sample_df() -> DADF:
    # Number of columns generated
    num_of_cols: int = random.randint(5, 10)
    # List of column name
    col_name: list = Generator.generate_string(
        Charset.LOWERCASE, unique=True, times=num_of_cols
    )
    # Create DataFrame
    df = pd.DataFrame(
        np.random.randn(random.randint(5, 100), num_of_cols), columns=col_name
    )
    out = DADF(df)
    return out


@pytest.fixture
def sample_df_2() -> DADF:
    return DADF.sample_df()


@pytest.fixture
def sample_df_3():
    sample = DADF.sample_df(size=SAMPLE_SIZE)
    sample["city"] = [x.city for x in sample_city_data]
    return sample


# MARK: test
class TestDataAnalystDataFrameBase: ...


class TestSplittedDF: ...


class TestDADF:
    """absfuyu.extra.da.dadf.DADF"""

    # Drop cols
    def test_drop_rightmost(self, sample_df: DADF) -> None:
        num_of_cols_drop = random.randint(1, 4)

        num_of_cols_current = sample_df.shape[1]
        sample_df.drop_rightmost(num_of_cols_drop)
        num_of_cols_modified = sample_df.shape[1]

        condition = (num_of_cols_current - num_of_cols_modified) == num_of_cols_drop
        assert condition

    # Add blank column
    def test_add_blank_column(self, sample_df: DADF) -> None:
        original_num_of_cols = sample_df.shape[1]
        sample_df.add_blank_column("new_col", 0)
        new_num_of_cols = sample_df.shape[1]

        condition = (new_num_of_cols - original_num_of_cols) == 1 and sum(
            sample_df["new_col"]
        ) == 0
        assert condition

    # Add date column
    def test_add_date_from_month(self, sample_df_2: DADF) -> None:
        sample_df_2.add_detail_date("date", mode="m")
        original_num_of_cols = sample_df_2.shape[1]
        sample_df_2.add_date_from_month("month", col_name="mod_date")
        new_num_of_cols = sample_df_2.shape[1]

        original_month = sample_df_2["month"][0]
        modified_month = sample_df_2["mod_date"][0].month

        # assert original_month == modified_month
        condition = (
            new_num_of_cols - original_num_of_cols
        ) == 1 and original_month == modified_month
        assert condition

    def test_add_date_column(self, sample_df_2: DADF) -> None:
        # Get random mode
        mode_list = ["d", "w", "m", "y"]
        test_mode = list(
            map(lambda x: "".join(x), Generator.combinations_range(mode_list))
        )
        random_mode = random.choice(test_mode)
        num_of_new_cols = len(random_mode)

        # Convert
        original_num_of_cols = sample_df_2.shape[1]
        sample_df_2.add_detail_date("date", mode=random_mode)
        new_num_of_cols = sample_df_2.shape[1]
        assert (new_num_of_cols - original_num_of_cols) == num_of_new_cols

    # Join and split
    def test_split_df(self, sample_df_2: DADF) -> None:
        test = sample_df_2.split_na("missing_value")
        assert len(test) > 1

    def test_split_df_2(self, sample_df_2: DADF) -> None:
        test = SplittedDF.divide_dataframe(sample_df_2, "number_range")
        assert len(test) > 1

    def test_join_df(self, sample_df_2: DADF) -> None:
        test = sample_df_2.split_na("missing_value")
        out = test.concat()
        assert out.shape[0] == 100

    def test_join_df_2(self, sample_df_2: DADF) -> None:
        """This test static method"""
        test = SplittedDF.divide_dataframe(sample_df_2, "number_range")
        out = SplittedDF.concat_df(test)
        assert out.shape[0] == 100

    # Threshold filter
    def test_threshold_filter(self, sample_df_2: DADF) -> None:
        original_num_of_cols = sample_df_2.shape[1]
        sample_df_2.threshold_filter("number_range", 11)
        new_num_of_cols = sample_df_2.shape[1]

        # Check new column
        assert (new_num_of_cols - original_num_of_cols) == 1

        # Check filler value
        test: list = sample_df_2["number_range_filtered"].unique().tolist()
        try:
            test.index("Other")
            assert True
        except Exception:
            pass

        # Check len
        test1 = sample_df_2["number_range"].unique().tolist()
        assert (len(test1) - len(test)) >= 1

    # Convert city
    def test_convert_city(self, sample_df_3: DADF) -> None:
        original_num_of_cols = sample_df_3.shape[1]
        sample_df_3.convert_city("city", city_list=sample_city_data)
        new_num_of_cols = sample_df_3.shape[1]
        assert (new_num_of_cols - original_num_of_cols) == 2

    # Diff
    def test_get_different_rows(self) -> None:
        df1 = DADF({"A": [1, 2, 3, 4], "B": [5, 6, 7, 8]})
        df2 = DADF({"A": [1, 2, 3, 4], "B": [7, 6, 6, 8]})
        assert df1.get_different_rows(df2).to_dict() == {
            "A": {0: 1, 2: 3},
            "B": {0: 7, 2: 6},
        }

    # Merge
    def test_merge_left(self) -> None:
        df1 = DADF(
            {
                "id": [1, 2, 5],
                "name": ["Alice", "Bob", "Rich"],
                "age": [20, 20, 20],
            }
        )
        df2 = DADF(
            {
                "id": [1, 2, 3],
                "age": [25, 30, 45],
                "department": ["HR", "IT", "PM"],
                "salary": [50000, 60000, 55000],
            }
        )
        assert df1.merge_left(df2, on="id").fillna("N/A").to_dict() == {
            "id": {0: 1, 1: 2, 2: 5},
            "name": {0: "Alice", 1: "Bob", 2: "Rich"},
            "age_x": {0: 20, 1: 20, 2: 20},
            "age_y": {0: 25.0, 1: 30.0, 2: "N/A"},
            "department": {0: "HR", 1: "IT", 2: "N/A"},
            "salary": {0: 50000.0, 1: 60000.0, 2: "N/A"},
        }
        assert df1.merge_left(df2, on="id", columns=["salary"]).fillna(
            "N/A"
        ).to_dict() == {
            "id": {0: 1, 1: 2, 2: 5},
            "name": {0: "Alice", 1: "Bob", 2: "Rich"},
            "age": {0: 25.0, 1: 30.0, 2: "N/A"},
            "department": {0: "HR", 1: "IT", 2: "N/A"},
            "salary": {0: 50000.0, 1: 60000.0, 2: "N/A"},
        }

    # Apply not null
    def test_apply_notnull(self) -> None:
        df = DADF({"A": [1, 2, 3], "B": [4, None, 6]})
        assert df.apply_notnull("B", lambda _: "Replaced").fillna("N/A").to_dict() == {
            "A": {0: 1, 1: 2, 2: 3},
            "B": {0: "Replaced", 1: "N/A", 2: "Replaced"},
        }

    def test_apply_notnull_row(self) -> None:
        df = DADF({"A": [None, 2, 3, 4], "B": [1, None, 3, 4], "C": [None, 2, None, 4]})
        assert df.apply_notnull_row().fillna("N/A").to_dict() == {
            "A": {0: "N/A", 1: 2.0, 2: 3.0, 3: 4.0},
            "B": {0: 1.0, 1: "N/A", 2: 3.0, 3: 4.0},
            "C": {0: "N/A", 1: 2.0, 2: "N/A", 3: 4.0},
            "applied_row_null": {0: False, 1: False, 2: False, 3: True},
        }
        df = DADF({"A": [None, 2, 3, 4], "B": [1, None, 3, 4], "C": [None, 2, None, 4]})
        assert df.apply_notnull_row(0, 1).fillna("N/A").to_dict() == {
            "A": {0: "N/A", 1: 2.0, 2: 3.0, 3: 4.0},
            "B": {0: 1.0, 1: "N/A", 2: 3.0, 3: 4.0},
            "C": {0: "N/A", 1: 2.0, 2: "N/A", 3: 4.0},
            "applied_row_null": {0: 0, 1: 0, 2: 0, 3: 1},
        }
        df = DADF({"A": [None, 2, 3, 4], "B": [1, None, 3, 4], "C": [None, 2, None, 4]})
        assert df.apply_notnull_row(
            lambda _: "n", lambda _: "y", col_name="mod"
        ).fillna("N/A").to_dict() == {
            "A": {0: "N/A", 1: 2.0, 2: 3.0, 3: 4.0},
            "B": {0: 1.0, 1: "N/A", 2: 3.0, 3: 4.0},
            "C": {0: "N/A", 1: 2.0, 2: "N/A", 3: 4.0},
            "mod": {0: "n", 1: "n", 2: "n", 3: "y"},
        }



"""
Test: Util

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

from pathlib import Path

import pytest

from absfuyu.config import CONFIG_PATH
from absfuyu.util import convert_to_raw_unicode, set_max, set_min, set_min_max
from absfuyu.util.json_method import JsonFile
from absfuyu.util.path import Directory


# MARK: util
@pytest.mark.abs_util
class TestUtil:
    """absfuyu.util"""

    @pytest.mark.skip
    @pytest.mark.parametrize(["value", "output"], [(10, 10), (-5, 0)])
    def test_set_min(self, value: int, output: int) -> None:
        assert set_min(value, min_value=0) == output

    @pytest.mark.skip
    @pytest.mark.parametrize(["value", "output"], [(200, 100), (10, 10)])
    def test_set_max(self, value: int, output: int) -> None:
        assert set_max(value, max_value=100) == output

    @pytest.mark.parametrize(["value", "output"], [(50, 50), (-10, 0), (200, 100)])
    def test_set_min_max(self, value: int, output: int) -> None:
        assert set_min_max(value, min_value=0, max_value=100) == output

    def test_convert_to_raw_unicode(self) -> None:
        assert convert_to_raw_unicode("résumé") == r"r\u00E9sum\u00E9"
        assert (
            convert_to_raw_unicode("résumé", partial=False)
            == r"\u0072\u00E9\u0073\u0075\u006D\u00E9"
        )


# MARK: json
@pytest.mark.abs_util
class TestUtilJsonMethod:
    """absfuyu.util.json_method"""

    def test_json_load(self) -> None:
        instance = JsonFile(CONFIG_PATH)
        loaded = instance.load_json()
        assert isinstance(loaded, dict)


# MARK: path
@pytest.mark.abs_util
class TestUtilPath:
    """absfuyu.util.path"""

    @pytest.mark.skip  # This takes too long
    def test_list_structure(self) -> None:
        instance = Directory(source_path=Path.cwd())
        assert instance.list_structure_pkg()



"""
Absfuyu: Core
-------------
Bases for other features (with library)

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    # Class
    "ShowAllMethodsMixinInspectVer",
    # Metaclass
    "PerformanceTrackingMeta",
    # Class decorator
    "positive_class_init_args",
]


# Library
# ---------------------------------------------------------------------------
import time
import tracemalloc
from collections.abc import Callable
from functools import wraps
from inspect import getmro, isfunction
from types import MethodType
from typing import Any, ParamSpec, TypeVar

# Type
# ---------------------------------------------------------------------------
P = ParamSpec("P")  # Parameter type
T = TypeVar("T", bound=type)  # Type type - Can be any subtype of `type`
# R = TypeVar("R")  # Return type


# Class
# ---------------------------------------------------------------------------
class ShowAllMethodsMixinInspectVer:
    @classmethod
    def show_all_methods(
        cls,
        include_classmethod: bool = True,
        classmethod_indicator: str = "<classmethod>",
    ) -> dict[str, list[str]]:
        result = {}
        for base in getmro(cls)[::-1]:
            methods = []
            # for name, attr in inspect.getmembers(base, predicate=callable):
            for name, attr in base.__dict__.items():
                if name.startswith("__"):
                    continue
                if isfunction(attr):
                    methods.append(name)
                # if inspect.ismethod(attr):
                if isinstance(attr, (classmethod, MethodType)) and include_classmethod:
                    methods.append(f"{name} {classmethod_indicator}")
            if methods:
                result[base.__name__] = sorted(methods)
        return result


# Metaclass
# ---------------------------------------------------------------------------
class PerformanceTrackingMeta(type):
    """
    A metaclass that tracks the instantiation time of classes.

    Usage:
    ------
    >>> class Demo(metaclass=PerformanceTrackingMeta):
    ...     def __init__(self): ...
    >>> Demo()
    --------------------------------------
    Class: Demo
    Memory usage:            0.000000 MB
    Peak memory usage:       0.000008 MB
    Time elapsed:            0.000001 s
    --------------------------------------
    """

    def __new__(mcs, name: str, bases: tuple, attrs: dict[str, Any]):
        """
        Intercepts class creation to wrap the ``__init__`` method.

        Parameters
        ----------
        mcs
            The metaclass

        name : str
            The class name

        bases : tuple
            Tuple of base classes

        attrs : dict[str, Any]
            Dictionary of attributes
        """
        if "__init__" in attrs:
            original_init = attrs["__init__"]

            @wraps(original_init)
            def wrapped_init(self, *args, **kwargs):
                # Performance check
                tracemalloc.start()
                start_time = time.perf_counter()

                # Run __init__()
                original_init(self, *args, **kwargs)

                # Performance stop
                current, peak = tracemalloc.get_traced_memory()
                end_time = time.perf_counter()
                tracemalloc.stop()
                creation_time = end_time - start_time

                # Print output
                print(
                    f"{''.ljust(38, '-')}\n"
                    f"Class: {name}\n"
                    f"Memory usage:\t\t {current / 10**6:,.6f} MB\n"
                    f"Peak memory usage:\t {peak / 10**6:,.6f} MB\n"
                    f"Time elapsed:\t\t {creation_time:,.6f} s\n"
                    f"{''.ljust(38, '-')}"
                )

            attrs["__init__"] = wrapped_init
        return super().__new__(mcs, name, bases, attrs)


# Decorator
# ---------------------------------------------------------------------------
def positive_class_init_args(cls: T):
    """
    A class decorator that ensures all arguments in the ``__init__()`` method are positive.
    """
    # original_init: Callable[P, None] | None = getattr(cls, "__init__", None)
    # if original_init is None:
    #     return cls
    try:
        original_init: Callable[P, None] = cls.__init__  # type: ignore
    except AttributeError:
        return cls

    @wraps(original_init)
    def new_init(self, *args: P.args, **kwargs: P.kwargs):
        # Check if all positional arguments are positive
        for arg in args:
            if isinstance(arg, (int, float)) and arg < 0:
                raise ValueError(f"Argument {arg} must be positive")

        # Check if all keyword arguments are positive
        for key, value in kwargs.items():
            if isinstance(value, (int, float)) and value < 0:
                raise ValueError(f"Argument {key}={value} must be positive")

        # Call the original __init__ method
        original_init(self, *args, **kwargs)

    # setattr(cls, "__init__", new_init)
    cls.__init__ = new_init  # type: ignore
    return cls



"""
Test: Data extension - Text

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.dxt import Text, TextAnalyzeDictFormat


@pytest.fixture
def example_long_text():
    return Text(
        "This is an extremely long text that even surpass my expectation and the rest of this text probably contains some useless stuff"
    )


@pytest.mark.abs_dxt
class TestText:
    """absfuyu.dxt.Text"""

    # analyze
    @pytest.mark.parametrize(
        ["value", "output"],
        [
            (
                "Lmao",
                {
                    "digit": 0,
                    "uppercase": 1,
                    "lowercase": 3,
                    "other": 0,
                },
            ),
            ("Lmao$$TEST.", {"digit": 0, "uppercase": 5, "lowercase": 3, "other": 3}),
        ],
    )
    def test_analyze(self, value: str, output: TextAnalyzeDictFormat) -> None:
        assert Text(value).analyze() == output

    # hex
    def test_to_hex(self) -> None:
        assert Text("Hello World").to_hex(raw=True) == "48656c6c6f20576f726c64"
        assert (
            Text("Hello World").to_hex()
            == "\\x48\\x65\\x6c\\x6c\\x6f\\x20\\x57\\x6f\\x72\\x6c\\x64"
        )

    # pangram
    @pytest.mark.parametrize(
        ["value", "output"],
        [
            ("abcdeFghijklmnopqrstuvwxyz", True),
            ("abcdefghijklmnOpqrstuvwxy", False),
            ("abcdeFghijklmnopqrstuvwxyzsdsd", True),
            ("abcdeFghijklmnopqrs tuvwxyzsdsd0", True),
        ],
    )
    def test_is_pangram(self, value: str, output: bool) -> None:
        assert Text(value).is_pangram() is output

    def test_is_pangram_custom(self) -> None:
        custom_alphabet = {"a", "b", "c"}
        assert Text("abc").is_pangram(custom_alphabet) is True
        assert Text("ab").is_pangram(custom_alphabet) is False
        assert Text("abcd").is_pangram(custom_alphabet) is True

    # palindrome
    @pytest.mark.parametrize(
        ["value", "output"], [("madam", True), ("racecar", True), ("bomb", False)]
    )
    def test_is_palindrome(self, value: str, output: bool) -> None:
        assert Text(value).is_palindrome() is output

    # reverse
    def test_reverse(self) -> None:
        assert Text("abc").reverse() == "cba"

    # random capslock
    @pytest.mark.parametrize("value", ["random", "capslock"])
    def test_random_capslock(self, value: str) -> None:
        test_0_percent: list[Text] = [
            Text(value).random_capslock(0) for _ in range(1000)
        ]
        test_50_percent: list[Text] = [
            Text(value).random_capslock(50) for _ in range(1000)
        ]
        test_100_percent: list[Text] = [
            Text(value).random_capslock(100) for _ in range(1000)
        ]
        assert len(list(set(test_0_percent))) == 1
        assert len(list(set(test_100_percent))) == 1
        assert Text(value).random_capslock(0) == value.lower()
        assert Text(value).random_capslock(100) == value.upper()

        try:
            assert len(list(set(test_50_percent))) != 1
            assert (
                Text(value).random_capslock(50) != value.lower()
                and Text(value).random_capslock(50) != value.upper()
            )
        except Exception as e:
            assert str(e)

    # divide
    def test_divide(self, example_long_text: Text) -> None:
        assert example_long_text.divide().__len__() == 3
        assert example_long_text.divide(string_split_size=10).__len__() == 13

    def test_divide_with_variable(self, example_long_text: Text) -> None:
        assert example_long_text.divide_with_variable(
            split_size=60, custom_var_name="abc"
        ) == [
            "abc1='This is an extremely long text that even surpass my expectat'",
            "abc2='ion and the rest of this text probably contains some useless'",
            "abc3=' stuff'",
            "abc=abc1+abc2+abc3",
            "abc",
        ]

    def test_divide_with_variable_2(self, example_long_text: Text) -> None:
        """Check for list len"""
        assert (
            example_long_text.divide_with_variable(
                split_size=60, custom_var_name="abc"
            ).__len__()
            == 5
        )

    # to list
    def test_to_list(self) -> None:
        assert isinstance(Text("test").to_list(), list)
        # assert isinstance(Text("test").to_listext(), ListExt)

    # count pattern
    @pytest.mark.parametrize("value", ["Test sentence"])
    @pytest.mark.parametrize(
        ["pattern", "output"],
        [("ten", 1), ("t", 2), ("a", 0)],
    )
    def test_count_pattern(self, value: str, pattern: str, output: int) -> None:
        assert Text(value).count_pattern(pattern) == output

    def test_count_pattern_2(self) -> None:
        assert Text("Test sentence").count_pattern("t", ignore_capslock=True) == 3

    def test_count_pattern_error(self) -> None:
        with pytest.raises(ValueError) as excinfo:
            Text("Test").count_pattern("tenss")
        assert str(excinfo.value)

    # hapax
    def test_hapax(self) -> None:
        assert Text("A a. a, b c c= C| d d").hapax() == [
            "a",
            "a.",
            "a,",
            "b",
            "c",
            "c=",
            "c|",
        ]

    def test_hapax_2(self) -> None:
        assert Text("A a. a, b c c= C| d d").hapax(strict=True) == ["b"]

    # shorten
    def test_shorten(self, example_long_text: Text) -> None:
        example_long_text.shorten()

    def test_shorten_negative_parameter(self, example_long_text: Text) -> None:
        example_long_text.shorten(-99)



"""
Absfuyu: Game
-------------
Contain some game that can be played on terminal

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "game_escapeLoop",
    "game_RockPaperScissors",
]


# Library
# ---------------------------------------------------------------------------
import random
import time

from .game_stat import GameStats

# Game
# ---------------------------------------------------------------------------
# Escape loop
_ESCAPE_LOOP_GAME_MSG = """\
Are you sure about this?
Don't leave me =((
I can't believe you did this to me!
Are you very much sure?
I'll be sad. Pick again please.
I still don't believe you.
Choose again.
You actually have to answer the correct keyword
I think you need to choose again.
Last chance!
Okay okay, i believe you ;)
Almost there.
I can do this all day
So close!
You can't escape from me.
How are you still here, just to suffer?
Never gonna give you up
Never gonna let you down
"""


def game_escapeLoop() -> None:
    """Try to escape the infinite loop"""

    init = True
    welcome_messages = [
        "Congrats! You are now stuck inside an infinite loop.",
        "Do you want to escape this loop?",
    ]

    num1 = random.choice(range(2, 13))
    num2 = random.choice(range(2, 13))
    hidden_answer = str(num1 * num2)

    gm_msg = {x for x in _ESCAPE_LOOP_GAME_MSG.splitlines() if len(x) > 0}
    game_messages = list(gm_msg) + [
        f"Hint 01: The keyword is: {num1}",
        f"Hint 02: {num2}",
    ]

    congrats_messages = ["Congratulation! You've escaped."]

    start_time = time.time()
    while True:
        # Welcome
        if init:
            for x in welcome_messages:
                print(x)
            answer = str(input())
            init = False

        # Random text
        mess = random.choice(game_messages)
        print(mess)

        # Condition check
        answer = str(input())
        if answer == hidden_answer:
            for x in congrats_messages:
                print(x)
            stop_time = time.time()
            break
    print(f"= Escaped in {stop_time - start_time:,.2f}s =")


# Rock Paper Scissors
def game_RockPaperScissors(hard_mode: bool = False) -> GameStats:
    """
    Rock Paper Scissors

    :param hard_mode: The bot only win or draw (Default: ``False``)
    :type hard_mode: bool
    """

    state_dict = {"0": "rock", "1": "paper", "2": "scissors"}

    init = True

    end_message = "end"

    welcome_messages = [
        "Welcome to Rock Paper Scissors",
        f"Type '{end_message}' to end",
    ]

    game_messages = [
        "Pick one option to begin:",
        " ".join([f"{k}={v}" for k, v in state_dict.items()]),
    ]

    game_summary = GameStats()

    while True:
        # Welcome
        if init:
            for x in welcome_messages:
                print(x)
            init = False

        # Game start
        print("")
        for x in game_messages:
            print(x)

        # Player's choice
        answer = input().strip().lower()

        # Condition check
        if answer == end_message:
            print(game_summary)
            break

        elif answer not in ["0", "1", "2"]:
            print("Invalid option. Choose again!")

        else:
            # Calculation
            if hard_mode:
                if answer == "0":
                    game_choice = random.choice(["0", "1"])
                if answer == "1":
                    game_choice = random.choice(["1", "2"])
                if answer == "2":
                    game_choice = random.choice(["0", "2"])
            else:
                game_choice = random.choice(["0", "1", "2"])
            print(f"You picked: {state_dict[answer]}")
            print(f"BOT picked: {state_dict[game_choice]}")

            if answer == "2" and game_choice == "0":
                print("You Lose!")
                game_summary.update_score("lose")
            elif answer == "0" and game_choice == "2":
                print("You Win!")
                game_summary.update_score("win")
            elif answer == game_choice:
                print("Draw Match!")
                game_summary.update_score("draw")
            elif answer > game_choice:
                print("You Win!")
                game_summary.update_score("win")
            else:
                print("You Lose!")
                game_summary.update_score("lose")

    return game_summary



"""
Absfuyu: Picture converter
--------------------------
Image converter


Version: 5.15.0
Date updated: 05/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["ImgConverter"]


# Library
# ---------------------------------------------------------------------------
import logging
import shutil
from functools import partial
from importlib.util import find_spec as check_for_package_installed
from pathlib import Path
from typing import Any, Literal, Protocol

from absfuyu.core.dummy_func import tqdm as tqdm_base
from absfuyu.util.path import DirectorySelectMixin

try:
    from PIL import Image
    from PIL import features as pil_features
    from PIL.ImageFile import ImageFile
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:  # type: ignore
        cmd = "python -m pip install -U absfuyu[pic]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[pic] package")  # noqa: B904
else:
    from PIL import Image
    from PIL import features as pil_features
    from PIL.ImageFile import ImageFile

try:
    from pillow_heif import register_heif_opener  # type: ignore
except ImportError:
    from absfuyu.core.dummy_func import dummy_function as register_heif_opener
register_heif_opener()


# Setup
# ---------------------------------------------------------------------------
tqdm = partial(tqdm_base, unit_scale=True, dynamic_ncols=True)
SupportedImageFormat = Literal[".jpg", ".jpeg", ".png", ".webp"]


# Exporter/Converter
# ---------------------------------------------------------------------------
class SupportImageConverter(Protocol):
    # Callable[[ImageFile, Path], None]
    def __call__(self, image: ImageFile, path: Path, **params: Any) -> None: ...


def _image_convert_default(image: ImageFile, path: Path, **params: Any) -> None:
    """
    Default convert image function

    Parameters
    ----------
    image : ImageFile
        Image file

    path : Path
        Path to export
    """
    image.save(path, **params)


def _image_convert_webp(image: ImageFile, path: Path, **params: Any) -> None:
    """
    Convert image to .webp format (with custom settings)

    Parameters
    ----------
    image : ImageFile
        Image file

    path : Path
        Path to export
    """
    image.save(
        path,
        format="WEBP",
        lossless=True,
        quality=100,
        alpha_quality=100,
        method=4,
        exact=False,  # If true, preserve the transparent RGB values. Otherwise, discard invisible RGB values for better compression. Defaults to false.
        **params,
    )


def _image_convert_png(image: ImageFile, path: Path, **params: Any) -> None:
    """
    Convert image to .png format (with custom settings)

    Parameters
    ----------
    image : ImageFile
        Image file

    path : Path
        Path to export
    """
    image.save(
        path,
        format="PNG",
        # optimize=True,
        compress_level=6,
        **params,
    )


def _image_convert_jpg(image: ImageFile, path: Path, **params: Any) -> None:
    """
    Convert image to .jpg format (with custom settings)

    Parameters
    ----------
    image : ImageFile
        Image file

    path : Path
        Path to export
    """

    if image.mode == "RGBA":
        white_background = Image.new("RGB", image.size, (255, 255, 255))
        white_background.paste(image, mask=image.getchannel("A"))
        image = white_background
    else:
        image = image.convert("RGB")

    image.save(
        path,
        format="JPEG",
        optimize=True,
        keep_rgb=True,
        **params,
    )


# Class
# ---------------------------------------------------------------------------
class ImgConverter(DirectorySelectMixin):
    _IMAGE_CONVERTER: dict[str, SupportImageConverter] = {
        "default": _image_convert_default,
        ".webp": _image_convert_webp,
        ".png": _image_convert_png,
        ".jpg": _image_convert_jpg,
        ".jpeg": _image_convert_jpg,
    }

    def __init__(
        self,
        source_path: str | Path,
        create_if_not_exist: bool = False,
        backup_dir_name: str | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        super().__init__(source_path, create_if_not_exist)

        # Supported image extension
        self._supported_image_format = [".png"]
        self._register_img_format()

        # Backup
        if backup_dir_name is None:
            backup_dir_name = "ZZZ_Backup"
        self.backup_path = self.source_path.joinpath(backup_dir_name)

        # Not available yet
        self.logger = logger

    # Extra format
    @classmethod
    def install_all_extension(cls) -> None:
        """
        Install all extra package to unlock all features
        """
        extra = [
            "pillow_heif",  # heic support
            "defusedxml",  # xmp
            "olefile",  # FPX and MIC images
        ]
        base = ["pip", "install", "-U"]
        base.extend(extra)

        import subprocess

        subprocess.run(base)

    def _register_img_format(self) -> None:
        """
        Try to register these format:
        - ``.webp``
        - ``.heif``, ``.heic`` (``pillow_heif`` package required)
        """
        if pil_features.check("jpg"):
            self._supported_image_format.extend([".jpg", ".jpeg"])
        if pil_features.check("webp"):
            self._supported_image_format.append(".webp")

        if check_for_package_installed("pillow_heif", "pillow_heif") is not None:
            self._supported_image_format.extend([".heic", ".heif"])

    @property
    def supported_image_format(self) -> list[str]:
        """
        Supported image format

        Returns
        -------
        list[str]
            Supported image format
        """
        return self._supported_image_format

    @classmethod
    def add_converter(cls, format_name: str, converter_func: SupportImageConverter) -> None:
        """
        Add image converter function to a format

        Parameters
        ----------
        format_name : str
            Image format name

        converter_func : SupportImageConverter
            Converter function


        Example:
        --------
        >>> ImgConverter.add_converter(".png", convert_to_png)
        """
        cls._IMAGE_CONVERTER[format_name] = converter_func

    # Support
    def _make_suffix_selection(self, exclude_suffix: str) -> tuple[str, ...]:
        """
        Make suffix selection (exclude the image with converted to suffix)

        Parameters
        ----------
        exclude_suffix : str
            Converted to suffix

        Returns
        -------
        tuple[str, ...]
            Suffix selection
        """
        # out = []
        # for x in self._supported_image_format:
        #     if x.lower() == exclude_suffix.lower():
        #         continue
        #     out.append(x.lower())
        #     out.append(x.upper())
        out = (x for x in self._supported_image_format if x.lower() != exclude_suffix.lower())
        return tuple(out)

    def _make_backup(self, src_file: Path) -> None:
        dest = self.backup_path.joinpath(src_file.name)
        shutil.move(src_file, dest)

    # Convert
    def _image_convert_legacy(
        self,
        path: Path,
        to_format: SupportedImageFormat | None = None,
        lossless: bool = True,
        compression_level: int | None = None,
    ) -> None:
        """
        Convert image to other format (settings are mostly for .webp format)

        Parameters
        ----------
        path : Path
            Path to image

        to_format : SupportedImageFormat | None, optional
            New image format, by default None

        lossless : bool, optional
            Lossless compression, by default True

        compression_level : int | None, optional
            Compression level, by default None
        """
        # Load image
        new_suffix = path.suffix if to_format is None else to_format
        image = Image.open(path)

        # Extract metadata
        # exif = image.info.get("exif")
        # xmp = image.getxmp()
        # icc_profile = image.info.get("icc_profile")
        xmp = image.info.get("xmp")
        exif = image.getexif()
        icc_profile = image.info.get("icc_profile")
        # print(image.info.keys())

        # Save
        image.save(
            path.with_suffix(new_suffix),
            format=new_suffix[1:].upper(),
            lossless=lossless,
            quality=100,
            alpha_quality=100,
            method=(4 if compression_level is None else compression_level),
            exact=False,  # If true, preserve the transparent RGB values. Otherwise, discard invisible RGB values for better compression. Defaults to false.
            exif=exif,
            icc_profile=icc_profile,
            xmp=xmp,
        )

    def _image_convert(
        self,
        path: Path,
        to_format: str | None = None,
    ) -> None:
        """
        Convert image to other format

        Parameters
        ----------
        path : Path
            Path to image

        to_format : SupportedImageFormat | None, optional
            New image format, by default None
        """
        # Load image
        new_suffix = path.suffix if to_format is None else to_format
        image = Image.open(path)

        # Extract metadata
        save_kwargs = {}
        if exif := image.getexif():
            save_kwargs["exif"] = exif
        if icc := image.info.get("icc_profile"):
            save_kwargs["icc_profile"] = icc
        if xmp := image.info.get("xmp"):
            save_kwargs["xmp"] = xmp

        # # Convert image mode
        # if image.mode not in ("RGB", "RGBA", "L"):
        #     image = image.convert("RGBA")

        # Save
        convert_func = self._IMAGE_CONVERTER.get(new_suffix, _image_convert_default)
        # self.logger.debug(f"Using {convert_func}")
        convert_func(image, path.with_suffix(new_suffix), **save_kwargs)

    def img_convert(self, to_format: SupportedImageFormat | str, backup: bool = True) -> None:
        """
        Convert images in directory to desire format

        Parameters
        ----------
        to_format : SupportedImageFormat
            Format to convert

        backup : bool
            Move pictures to a backup folder

        Raises
        ------
        NotImplementedError
            Not supported image format
        """
        if to_format not in self._supported_image_format:
            raise NotImplementedError("Format not supported")

        imgs = self.select_all(*self._make_suffix_selection(to_format))

        for x in tqdm(imgs, desc=f"Converting to {to_format}"):
            try:
                self._image_convert(x, to_format=to_format)

                if backup:
                    self.backup_path.mkdir(parents=True, exist_ok=True)
                    self._make_backup(x)
            except TypeError as err:
                print(f" TYPE ERROR: {x} - {err}")
            except Exception as err:
                print(f" ERROR: {x} - {err}")



"""
Absfuyu: Obfuscator
-------------------
Obfuscate code

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["Obfuscator", "StrShifter"]


# Library
# ---------------------------------------------------------------------------
import base64
import codecs
import random
import zlib
from collections import deque
from string import Template
from typing import ClassVar

from absfuyu.core.baseclass import BaseClass, GetClassMembersMixin
from absfuyu.core.docstring import versionadded
from absfuyu.dxt import Text
from absfuyu.logger import logger
from absfuyu.tools.generator import Charset, Generator


# Class
# ---------------------------------------------------------------------------
@versionadded("5.0.0")
class StrShifter(BaseClass):
    """
    Shift characters in a string by a specified number of positions.

    Parameters
    ----------
    str_to_shift : str
        The string whose characters will be shifted.

    shift_by : int, optional
        The number of positions to shift the characters, by default ``5``.
    """

    __slots__ = ("_str_to_shift", "shift_by")

    def __init__(self, str_to_shift: str, shift_by: int = 5) -> None:
        """
        Initialize the StrShifter with the string to shift and the shift amount.

        Parameters
        ----------
        str_to_shift : str
            The string whose characters will be shifted.

        shift_by : int, optional
            The number of positions to shift the characters, by default ``5``.
        """
        if not isinstance(str_to_shift, str):
            raise TypeError("Value must be an instance of str")
        self._str_to_shift = str_to_shift
        self.shift_by = shift_by

    def _make_convert_table(self) -> dict[str, str]:
        """
        Create a translation table for shifting characters.

        Returns
        -------
        dict[str, str]
            A dictionary mapping each character to its shifted counterpart.
        """
        data = self._str_to_shift  # Make a copy

        unique_char_sorted = deque(sorted(list(set(data))))
        translate = unique_char_sorted.copy()
        translate.rotate(self.shift_by)
        convert_table = dict(zip(unique_char_sorted, translate))

        return convert_table

    def _use_convert_table(self, convert_table: dict[str, str]) -> str:
        """
        Convert the original string using the provided conversion table.

        Parameters
        ----------
        convert_table : dict[str, str]
            The conversion table mapping original characters to shifted characters.

        Returns
        -------
        str
            The transformed string after applying the conversion table.
        """
        return "".join([convert_table[char] for char in list(self._str_to_shift)])

    def shift(self) -> str:
        """
        Shift the characters in the string and return the new string.

        Returns
        -------
        str
            The resulting string after shifting.
        """
        return self._use_convert_table(self._make_convert_table())


class Obfuscator(GetClassMembersMixin):
    """
    Obfuscate code

    Parameters
    ----------
    code : str
        Code text

    base64_only : bool, optional
        - ``True``: encode in base64 form only
        - ``False``: base64, compress, rot13 (default)

    split_every : int, optional
        Split the long line of code every ``x`` character.
        Minimum is ``1``, by default ``60``

    variable_length : int, optional
        Length of variable name (when data string split).
        Minimum is ``7``, by default ``12``

    fake_data : bool, optional
        Generate additional meaningless data, by default ``False``
    """

    # Var
    LIB_BASE64_ONLY: ClassVar[list[str]] = ["base64"]
    LIB_FULL: ClassVar[list[str]] = ["base64", "codecs", "zlib"]

    # Template
    SINGLE_LINE_TEMPLATE: ClassVar[Template] = Template(
        "exec(bytes.fromhex('$one_line_code').decode('utf-8'))"
    )
    PRE_HEX_B64_TEMPLATE: ClassVar[Template] = Template(
        "eval(compile(base64.b64decode($encoded_string),$type_var,$execute))"
    )
    PRE_HEX_FULL_TEMPLATE: ClassVar[Template] = Template(
        "eval(compile(base64.b64decode(zlib.decompress(base64.b64decode(codecs."
        "encode($encoded_string,$codec_to_decode).encode()))),$type_var,$execute))"
    )

    def __init__(
        self,
        code: str,
        *,
        base64_only: bool = False,
        split_every: int = 60,
        variable_length: int = 12,
        fake_data: bool = False,
    ) -> None:
        """
        Obfuscator

        Parameters
        ----------
        code : str
            Code text

        base64_only : bool, optional
            - ``True``: encode in base64 form only
            - ``False``: base64, compress, rot13 (default)

        split_every : int, optional
            Split the long line of code every ``x`` character.
            Minimum is ``1``, by default ``60``

        variable_length : int, optional
            Length of variable name (when data string split).
            Minimum is ``7``, by default ``12``

        fake_data : bool, optional
            Generate additional meaningless data, by default ``False``
        """
        self.base_code = code
        self.base64_only = base64_only
        self.split_every_length = max(1, split_every)
        self.variable_length = max(7, variable_length)
        self.fake_data = fake_data

        # Setting
        self._library_import_variable_length = self.variable_length - 1
        self._splited_variable_length = self.variable_length
        self._decode_variable_length = self.variable_length + 3

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        repr_out_dict = {
            "base_code": "...",
            "base64_only": self.base64_only,
            "split_every_length": self.split_every_length,
            "variable_length": self.variable_length,
            "fake_data": self.fake_data,
        }
        repr_out = ", ".join([f"{k}={repr(v)}" for k, v in repr_out_dict.items()])
        return f"{self.__class__.__name__}({repr_out})"

    def to_single_line(self) -> str:
        """
        Convert multiple lines of code into one line

        Returns
        -------
        str
            Converted code.
        """
        newcode = self.base_code.encode("utf-8").hex()
        output = self.SINGLE_LINE_TEMPLATE.substitute(one_line_code=newcode)
        return output

    # Obfuscate original code
    def _obfuscate(self) -> str:
        """
        Convert multiple lines of code through multiple transformation
        (base64 -> compress -> base64 -> caesar (13))
        """
        code = self.base_code  # Make a copy
        logger.debug("Encoding...")

        b64_encode = base64.b64encode(code.encode())
        if self.base64_only:
            output = b64_encode.decode()
        else:
            compressed_data = zlib.compress(b64_encode)
            logger.debug(
                f"Compressed data: {str(compressed_data)} | Len: {len(str(compressed_data))}"
            )
            b64_encode_2 = base64.b64encode(compressed_data).decode()
            logger.debug(f"Base64 encode 2: {b64_encode_2} | Len: {len(b64_encode_2)}")
            caesar_data = codecs.encode(b64_encode_2, "rot_13")
            output = caesar_data

        logger.debug(f"Output: {output}")
        logger.debug("Code encoded.")
        return output

    @staticmethod
    def _convert_to_base64_decode(text: str, raw: bool = False) -> str:
        """
        Convert text into base64 and then return a code that decode that base64 code

        Parameters
        ----------
        text : str
            Code that need to convert

        raw : bool
            Return hex form only, by default ``False``
        """
        b64_encode_codec = base64.b64encode(text.encode()).decode()
        b64_decode_codec = f"base64.b64decode('{b64_encode_codec}'.encode()).decode()"
        hex = Text(b64_decode_codec).to_hex()
        out = f"eval('{hex}')"
        if raw:
            return hex
        return out

    # Generate output (decode obfuscated code)
    def _make_output_lib(self) -> list[str]:
        """Obfuscate the `import <lib>`"""
        output = []

        # Make import lib
        library_list = self.LIB_BASE64_ONLY if self.base64_only else self.LIB_FULL
        imports = [f"import {lib}" for lib in library_list]
        logger.debug(f"Lib: {imports}")

        # Convert to hex
        lib_hex = Text("\n".join(imports)).to_hex()
        output.append(f"exec('{lib_hex}')")
        logger.debug(f"Current output (import library): {output}")

        return output

    def _make_prep_for_decode_var(self) -> tuple[list[str], list[str]]:
        """
        ``<var> = "rot_13"``
        ``<var> = "<string>"``
        ``<var> = "exec"``

        Returns
        -------
        tuple[list[str], list[str]]
            - tuple[0]: output
            - tuple[1]: decode var name
        """
        output = []

        # Make variables for "rot_13", "<string>", "exec"
        dc_name_lst: list[str] = Generator.generate_string(
            charset=Charset.ALPHABET,
            size=self._decode_variable_length,
            times=3,
            unique=True,
        )

        # Assign and convert to hex
        encode_codec = "rot_13"  # full
        if not self.base64_only:  # full
            hex_0 = self._convert_to_base64_decode(encode_codec)
            output.append(f"{dc_name_lst[0]}={hex_0}")

        for i, x in enumerate(["<string>", "exec"], start=1):
            # hex_str = Text(x).to_hex()
            hex_str = self._convert_to_base64_decode(x)
            output.append(f"{dc_name_lst[i]}={hex_str}")
        logger.debug(f"Current output (decode variables): {output}")

        return output, dc_name_lst

    def _make_fake_output(self, input_size: int) -> list[str]:
        """Fake data"""
        output = []

        f1 = Generator.generate_string(
            charset=Charset.DEFAULT,
            size=input_size,
            times=1,
            string_type_if_1=True,
        )  # Generate fake data with len of original data
        f2 = Text(f1).divide_with_variable(
            self.split_every_length, self._splited_variable_length
        )
        output.extend(f2[:-1])

        # Random data
        bait_lst = Generator.generate_string(
            charset=Charset.ALPHABET, size=self._splited_variable_length, times=25
        )
        for x in bait_lst:
            output.append(
                f"{x}='{Generator.generate_string(charset=Charset.DEFAULT, size=self.split_every_length, times=1, string_type_if_1=True)}'"
            )

        random_eval_text = str(random.randint(1, 100))
        for _ in range(random.randint(10, 50)):
            random_eval_text += f"+{random.randint(1, 100)}"
        random_eval_text_final = Text(random_eval_text).to_hex()
        output.append(f"eval('{random_eval_text_final}')")

        return output

    def _make_obfuscate_output(self) -> list[str]:
        """
        Convert multiple lines of code through multiple transformation
        (base64 -> compress -> base64 -> caesar (13))

        Then return a list (obfuscated code) that can
        be print or export into .txt file
        """
        # Obfuscated code
        input_str = Text(self._obfuscate())

        # Generate output
        output = []

        # Import library
        output.extend(self._make_output_lib())

        # Append divided long text list
        input_list = input_str.divide_with_variable(
            split_size=self.split_every_length,
            split_var_len=self._splited_variable_length,
        )
        encoded_str = input_list[-1]  # Main var name that will later be used
        output.extend(input_list[:-1])  # Append list minus the last element
        logger.debug(f"Current output (encoded code): {output}")

        # Decode: encoded_str
        dc_out, dc_name_lst = self._make_prep_for_decode_var()
        output.extend(dc_out)

        if self.base64_only:  # b64
            pre_hex = self.PRE_HEX_B64_TEMPLATE.substitute(
                encoded_string=encoded_str,
                type_var=dc_name_lst[1],
                execute=dc_name_lst[2],
            )
        else:  # full
            pre_hex = self.PRE_HEX_FULL_TEMPLATE.substitute(
                encoded_string=encoded_str,
                codec_to_decode=dc_name_lst[0],
                type_var=dc_name_lst[1],
                execute=dc_name_lst[2],
            )

        t_hex = Text(pre_hex).to_hex()
        output.append(f"exec('{t_hex}')")
        logger.debug(f"Current output (decode code): {output}")

        # Fake data
        if self.fake_data:
            output.extend(self._make_fake_output(len(input_str)))

        logger.debug("Code obfuscated.")
        return output

    def obfuscate(self) -> str:
        """
        Obfuscate code

        Returns
        -------
        str
            Obfuscated code
        """
        return "\n".join(self._make_obfuscate_output())



"""
Absfuyu: Data Analysis [W.I.P]
------------------------------
Extension for ``pd.DataFrame``
(deprecated)

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Library
# ---------------------------------------------------------------------------
from absfuyu.extra.da.dadf import DADF  # noqa
from absfuyu.extra.da.df_func import (  # noqa
    compare_2_list,
    equalize_df,
    rename_with_dict,
)

# Class - DA
# TODO: drop dups | Combine: row with data, row NaN



"""
Absfuyu: Performance
--------------------
Performance Check

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)

Feature:
--------
- measure_performance
- function_debug
- retry
- var_check
- Checker
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # Wrapper
    "function_debug",
    "function_benchmark",
    "measure_performance",
    "retry",
    # Class
    "Checker",
]


# Library
# ---------------------------------------------------------------------------
import time
import tracemalloc
from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from inspect import getsource
from typing import Any, Literal, ParamSpec, TypeVar, overload

from absfuyu.core import deprecated, versionadded, versionchanged

# Type
# ---------------------------------------------------------------------------
P = ParamSpec("P")  # Parameter type
R = TypeVar("R")  # Return type


# Support
# ---------------------------------------------------------------------------
@dataclass
class BenchmarkResult:
    """
    Use ``format(BenchmarkResult(...), "seconds")`` to view result in seconds.
    """

    min_: float
    max_: float
    avg: float

    def __format__(self, format_spec: str) -> str:
        clsname = self.__class__.__name__
        if format_spec.lower().strip().startswith("seconds"):
            fields = [f"{x}={getattr(self, x):,.6f}s" for x in self._get_fields()]
            return f"{clsname}({', '.join(fields)})"
        return repr(self)

    @classmethod
    def _get_fields(cls) -> tuple[str, ...]:
        return tuple(cls.__dataclass_fields__)


# Function
# ---------------------------------------------------------------------------
@versionchanged("3.2.0", reason="Updated functionality")
def measure_performance(f: Callable[P, R]) -> Callable[P, R]:
    r"""
    Measure performance of a function

    Parameters
    ----------
    f : Callable [P, R]
        Function to measure

    Returns
    -------
    Callable
        A decorated function


    Usage
    -----
    Use this as a decorator (``@measure_performance``)

    Example:
    --------
    >>> @measure_performance
    >>> def test():
    ...     return 1 + 1
    >>> test()
        --------------------------------------
        Function: test
        Memory usage:		 0.000000 MB
        Peak memory usage:	 0.000000 MB
        Time elapsed (seconds):	 0.000002
        --------------------------------------

    """

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        # Performance check
        tracemalloc.start()  # Start memory measure
        start_time = time.perf_counter()  # Start time measure
        output = f(*args, **kwargs)  # Run function and save result into a variable
        current, peak = tracemalloc.get_traced_memory()  # Get memory stats
        finish_time = time.perf_counter()  # Get finished time
        tracemalloc.stop()  # End memory measure

        # Print output
        print(
            f"{'-' * 38}\n"
            f"Function: {f.__name__}\n"
            f"Memory usage:\t\t {current / 10**6:,.6f} MB\n"
            f"Peak memory usage:\t {peak / 10**6:,.6f} MB\n"
            f"Time elapsed (seconds):\t {finish_time - start_time:,.6f}\n"
            f"{'-' * 38}"
        )

        # Return
        return output

    return wrapper


@overload
def function_benchmark(func: Callable[P, R], /) -> Callable[P, R]: ...
@overload
def function_benchmark(*, n: int = 1) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def function_benchmark(
    *, n: int = 1, result_only: Literal[False] = False
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def function_benchmark(
    *, n: int = 1, result_only: Literal[True] = ...
) -> Callable[[Callable[P, R]], Callable[P, BenchmarkResult]]: ...


@versionadded("5.2.0")
def function_benchmark(
    func: Callable[P, R] | None = None, /, *, n: int = 1, result_only: bool = False
):
    """
    This run function for ``n`` times and calculate min, max, average runtime.

    Parameters
    ----------
    func : Callable[P, R] | None, optional
        Callable with parameter **P and returns R, by default ``None``

    n : int, optional
        Run how many times, by default ``1``

    result_only : bool, optional
        Returns BenchmarkResult instead of ``func`` result, by default ``False``


    Usage
    -----
    Use this as a decorator (``@function_benchmark``)

    Example:
    --------
    >>> @function_benchmark
    >>> def test():
    ...     return 1 + 1
    >>> test()
    BenchmarkResult(min_=0.000000s, max_=0.000000s, avg=0.000000s)
    2

    >>> @function_benchmark(n=1)
    >>> def test():
    ...     return 1 + 1
    >>> test()
    BenchmarkResult(min_=0.000000s, max_=0.000000s, avg=0.000000s)
    2
    """

    times = max(n, 1)

    def decorator(f: Callable[P, R]) -> Callable[P, R | BenchmarkResult]:
        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R | BenchmarkResult:
            output = f(*args, **kwargs)  # Run function and save result into a variable

            def _run() -> float:
                # Performance check
                start_time = time.perf_counter()  # Start time measure
                f(*args, **kwargs)
                finish_time = time.perf_counter()  # Get finished time
                return finish_time - start_time

            # run = (_run() for _ in range(times))
            run = [_run() for _ in range(times)]
            try:
                avg_runtime = sum(run) / len(run)
            except ZeroDivisionError:
                avg_runtime = min(run)
            result = BenchmarkResult(min(run), max(run), avg_runtime)

            if result_only:
                return result

            print(format(result, "seconds"))

            return output

        return wrapper

    if func is None:
        return decorator
    return decorator(func)


@versionadded("3.2.0")
def function_debug(f: Callable[P, R]) -> Callable[P, R]:
    """
    Print the function signature and return value

    Parameters
    ----------
    func : Callable
        Function to debug

    Returns
    -------
    Callable
        A decorated function


    Usage
    -----
    Use this as a decorator (``@function_debug``)

    Example:
    --------
    >>> @function_debug
    >>> def test(a: int, b: int):
    ...     return a + b
    >>> test(6, 8)
    Calling test(6, 8)
    test() returned 14
    """

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        # Get all parameters inputed
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={repr(v)}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)

        # Output
        print(f"Calling {f.__name__}({signature})")
        # logger.debug(f"Calling {func.__name__}({signature})")
        value = f(*args, **kwargs)
        print(f"{f.__name__}() returned {repr(value)}")
        # logger.debug(f"{func.__name__}() returned {repr(value)}")
        return value

    return wrapper


@versionadded("3.2.0")
def retry(retries: int, delay: float = 1):
    """
    Attempt to call a function, if it fails, try again with a specified delay.

    Parameters
    ----------
    retries : int
        The max amount of retries you want for the function call

    delay : int
        The delay (in seconds) between each function retry

    Returns
    -------
    Callable
        A decorated function


    Usage
    -----
    Use this as a decorator (``@retry``)

    Example:
    --------
    >>> @retry(retries=3, delay=1)
    >>> def test() -> None:
    ...     time.sleep(1)
    ...     raise Exception("Function error")
    >>> test()
    Running (1): test()
    Error: Exception('Function error') -> Retrying...
    Running (2): test()
    Error: Exception('Function error') -> Retrying...
    Running (3): test()
    Error: Exception('Function error').
    "test()" failed after 3 retries.
    """

    # Condition
    if retries < 1 or delay <= 0:
        raise ValueError("retries must be >= 1, delay must be >= 0")

    def decorator(f: Callable[P, R]) -> Callable[P, R]:
        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs):
            for i in range(1, retries + 1):
                try:
                    print(f"Running ({i}): {f.__name__}()")
                    return f(*args, **kwargs)
                except Exception as e:
                    # Break out of the loop if the max amount of retries is exceeded
                    if i == retries:
                        print(f"Error: {repr(e)}.")
                        print(f'"{f.__name__}()" failed after {retries} retries.')
                        break
                    else:
                        print(f"Error: {repr(e)} -> Retrying...")
                        time.sleep(
                            delay
                        )  # Add a delay before running the next iteration

        return wrapper

    return decorator


# Class
# ---------------------------------------------------------------------------
@deprecated("5.1.0", reason="Use `absfuyu.tools.inspector` instead")
class Checker:
    """
    Check a variable

    Parameters
    ----------
    variable : Any
        Variable that needed to check


    Example:
    --------
    >>> test = "test"
    >>> Checker(test).check()
    {'name': None, 'value': 'test', 'class': <class 'str'>, 'id': ...}
    """

    def __init__(self, variable: Any) -> None:
        self.item_to_check = variable

    def __str__(self) -> str:
        return self.item_to_check.__str__()  # type: ignore

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.item_to_check})"

    @property
    def name(self) -> Any | None:
        """``__name__`` of variable (if any)"""
        try:
            return self.item_to_check.__name__
        except Exception:
            return None

    @property
    def value(self) -> Any:
        """Value of the variable"""
        return self.item_to_check

    @property
    def docstring(self) -> str | None:
        """``__doc__`` of variable (if any)"""
        return self.item_to_check.__doc__  # type: ignore

    @property
    def class_(self) -> Any:
        """``class()`` of variable"""
        return type(self.item_to_check)

    @property
    def id_(self) -> int:
        """``id()`` of variable"""
        return id(self.item_to_check)

    @property
    def dir_(self) -> list[str]:
        """``dir()`` of variable"""
        # return self.item_to_check.__dir__()
        return [x for x in dir(self.item_to_check) if not x.startswith("__")]

    @property
    def source(self) -> str | None:
        """Source code of variable (if available)"""
        try:
            return getsource(self.item_to_check)
        except Exception:
            return None

    def check(self, full: bool = False) -> dict[str, Any]:
        """
        Check the variable

        Parameters
        ----------
        full : bool
            | ``True``: Shows full detail
            | ``False``: Hides ``dir``, ``docstring`` and source code
            | Default: ``False``

        Returns
        -------
        dict[str, Any]
            Check result
        """
        out = {
            "name": self.name,
            "value": self.value,
            "class": self.class_,
            "id": self.id_,
        }
        if full:
            out["dir"] = self.dir_
            out["docs"] = self.docstring
            out["source"] = self.source
        return out



"""
Absfuyu: Tools
--------------
Some useful tools

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    #     # Main
    "Checksum",
    "B64",
    "T2C",
    #     "Charset",
    #     "Generator",
    "Inspector",
    "inspect_all",
    #     "Obfuscator",
    #     "StrShifter",
]


# Library
# ---------------------------------------------------------------------------
from absfuyu.tools.checksum import Checksum
from absfuyu.tools.converter import Base64EncodeDecode as B64
from absfuyu.tools.converter import Text2Chemistry as T2C
from absfuyu.tools.inspector import Inspector, inspect_all

# from absfuyu.tools.generator import Charset, Generator  # circular import bug
# from absfuyu.tools.obfuscator import Obfuscator, StrShifter  # circular import bug



"""
Absfuyu: Passwordlib
--------------------
Password library

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["PasswordGenerator", "PasswordHash", "TOTP"]


# Library
# ---------------------------------------------------------------------------
import hashlib
import os
import random
import re
from typing import ClassVar, Literal, NamedTuple
from urllib.parse import quote, urlencode

from absfuyu.core.baseclass import BaseClass
from absfuyu.core.docstring import deprecated, versionadded
from absfuyu.dxt import DictExt, Text
from absfuyu.logger import logger
from absfuyu.pkg_data import DataList, DataLoader
from absfuyu.tools.generator import Charset, Generator


# Function
# ---------------------------------------------------------------------------
@deprecated("5.0.0")
@versionadded("4.2.0")
def _password_check(password: str) -> bool:
    """
    Verify the strength of ``password``.
    A password is considered strong if:

    - 8 characters length or more
    - 1 digit or more
    - 1 symbol or more
    - 1 uppercase letter or more
    - 1 lowercase letter or more

    :param password: Password want to be checked
    :type password: str
    :rtype: bool
    """

    # calculating the length
    length_error = len(password) < 8

    # searching for digits
    digit_error = re.search(r"\d", password) is None

    # searching for uppercase
    uppercase_error = re.search(r"[A-Z]", password) is None

    # searching for lowercase
    lowercase_error = re.search(r"[a-z]", password) is None

    # searching for symbols
    symbols = re.compile(r"[ !#$%&'()*+,-./[\\\]^_`{|}~" + r'"]')
    symbol_error = symbols.search(password) is None

    detail = {
        "length_error": length_error,
        "digit_error": digit_error,
        "uppercase_error": uppercase_error,
        "lowercase_error": lowercase_error,
        "symbol_error": symbol_error,
    }
    logger.debug(f"Password error summary: {detail}")

    return not any(
        [
            length_error,
            digit_error,
            uppercase_error,
            lowercase_error,
            symbol_error,
        ]
    )


# Class
# ---------------------------------------------------------------------------
class PasswordHash(NamedTuple):
    """
    Password hash

    Parameters
    ----------
    salt : bytes
        Salt

    key : bytes
        Key
    """

    salt: bytes
    key: bytes


@versionadded("4.2.0")
class PasswordGenerator(BaseClass):
    """Password Generator"""

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    @staticmethod
    def password_hash(password: str) -> PasswordHash:
        """
        Generate hash for password.

        Parameters
        ----------
        password : str
            Password string

        Returns
        -------
        PasswordHash
            Password hash contains salt and key
        """
        salt = os.urandom(32)
        key = hashlib.pbkdf2_hmac(
            hash_name="sha256",
            password=password.encode("utf-8"),
            salt=salt,
            iterations=100000,
        )
        out = PasswordHash(salt, key)
        return out

    @staticmethod
    def password_check(password: str) -> dict:
        """
        Check password's characteristic.

        Parameters
        ----------
        password : str
            Password string

        Returns
        -------
        dict
            Password's characteristic.
        """
        data = Text(password).analyze()
        data = DictExt(data).apply(lambda x: True if x > 0 else False)  # type: ignore
        data.__setitem__("length", len(password))
        return dict(data)

    # Password generator
    @staticmethod
    def generate_password(
        length: int = 8,
        include_uppercase: bool = True,
        include_number: bool = True,
        include_special: bool = True,
    ) -> str:
        r"""
        Generate a random password.

        Parameters
        ----------
        length : int
            | Length of the password.
            | Minimum value: ``8``
            | (Default: ``8``)

        include_uppercase : bool
            Include uppercase character in the password, by default ``True``

        include_number : bool
            Include digit character in the password, by default ``True``

        include_special : bool
            Include special character in the password, by default ``True``

        Returns
        -------
        str
            Generated password


        Example:
        --------
        >>> Password.generate_password()
        [T&b@mq2
        """
        charset = Charset.LOWERCASE
        check = 0

        if include_uppercase:
            charset += Charset.UPPERCASE
            check += 1

        if include_number:
            charset += Charset.DIGIT
            check += 1

        if include_special:
            charset += r"[!#$%&'()*+,-./]^_`{|}~\""
            check += 1

        while True:
            pwd = Generator.generate_string(
                charset=charset,
                size=max(length, 8),  # type: ignore
                times=1,
                string_type_if_1=True,
            )

            analyze = Text(pwd).analyze()  # Count each type of char

            s = sum([1 for x in analyze.values() if x > 0])  # type: ignore
            if s > check:  # Break loop if each type of char has atleast 1
                break
        return pwd  # type: ignore

    @staticmethod
    def generate_passphrase(
        num_of_blocks: int = 5,
        block_divider: str | None = None,
        first_letter_cap: bool = True,
        include_number: bool = True,
        *,
        custom_word_list: list[str] | None = None,
    ) -> str:
        """
        Generate a random passphrase

        Parameters
        ----------
        num_of_blocks : int
            Number of word used, by default ``5``

        block_divider : str
            Character symbol that between each word, by default ``"-"``

        first_letter_cap : bool
            Capitalize first character of each word, by default ``True``

        include_number : bool
            Add number to the end of each word, by default ``True``

        custom_word_list : list[str] | None
            Custom word list for passphrase generation, by default uses a list of 360K+ words

        Returns
        -------
        str
            Generated passphrase


        Example:
        --------
        >>> print(Password().generate_passphrase())
        Myomectomies7-Sully4-Torpedomen7-Netful2-Begaud8
        """
        words: list[str] = (
            DataLoader(DataList.PASSWORDLIB).load().decode().split(",")
            if not custom_word_list
            else custom_word_list
        )

        if block_divider is None:
            block_divider = "-"

        dat = [random.choice(words) for _ in range(num_of_blocks)]

        if first_letter_cap:
            dat = list(map(lambda x: x.title(), dat))

        if include_number:
            idx = random.choice(range(num_of_blocks))
            dat[idx] += str(random.choice(range(10)))

        return block_divider.join(dat)


@versionadded("5.0.0")
class TOTP(BaseClass):
    """
    A class to represent a Time-based One-Time Password (TOTP) generator.

    Parameters
    ----------
    secret : str
        The shared secret key used to generate the TOTP.

    name : str, optional
        | The name associated with the TOTP.
        | If not provided, by default ``"None"``.

    issuer : str, optional
        The issuer of the TOTP.

    algorithm : Literal["SHA1", "SHA256", "SHA512"], optional
        | The hashing algorithm used to generate the TOTP.
        | Must be one of ``"SHA1"``, ``"SHA256"``, or ``"SHA512"``.
        | By default ``"SHA1"``.

    digit : int, optional
        | The number of digits in the generated TOTP.
        | Must be greater than 0.
        | By default ``6``.

    period : int, optional
        | The time step in seconds for TOTP generation.
        | Must be greater than 0.
        | by default ``30``.
    """

    URL_SCHEME: ClassVar[str] = "otpauth://totp/"

    def __init__(
        self,
        secret: str,
        name: str | None = None,
        issuer: str | None = None,
        algorithm: Literal["SHA1", "SHA256", "SHA512"] = "SHA1",
        digit: int = 6,
        period: int = 30,
    ) -> None:
        """
        Initializes a TOTP instance.

        Parameters
        ----------
        secret : str
            The shared secret key used to generate the TOTP.

        name : str, optional
            | The name associated with the TOTP.
            | If not provided, by default ``"None"``.

        issuer : str, optional
            The issuer of the TOTP.

        algorithm : Literal["SHA1", "SHA256", "SHA512"], optional
            | The hashing algorithm used to generate the TOTP.
            | Must be one of ``"SHA1"``, ``"SHA256"``, or ``"SHA512"``.
            | By default ``"SHA1"``.

        digit : int, optional
            | The number of digits in the generated TOTP.
            | Must be greater than 0.
            | By default ``6``.

        period : int, optional
            | The time step in seconds for TOTP generation.
            | Must be greater than 0.
            | by default ``30``.
        """
        self.secret = secret.upper()
        self.name = name if name else "None"
        self.issuer = issuer
        self.algorithm = algorithm.upper()
        self.digit = max(digit, 1)  # digit must be larger than 0
        self.period = max(period, 1)  # period must be larger than 0

    def to_url(self) -> str:
        """
        Generates a URL for the TOTP in the otpauth format.

        The URL format is as follows:
            ``otpauth://totp/<name>?secret=<secret>&issuer=<issuer>&algorithm=<algorithm>&digit=<digit>&period=<period>``

        Returns
        -------
        str
            A URL representing the TOTP in otpauth format.
        """
        params = {
            "secret": self.secret,
            "issuer": self.issuer,
            "algorithm": self.algorithm,
            "digit": self.digit,
            "period": self.period,
        }
        # Filter out None values from the params dictionary
        filtered_params = {k: v for k, v in params.items() if v is not None}
        # filtered_params = {k: v for k, v in self.__dict__.items() if v is not None}

        name = quote(self.name)
        tail = urlencode(filtered_params, quote_via=quote)
        return f"{self.URL_SCHEME}{name}?{tail}"



"""
Absfuyu: Data Extension
-----------------------
dict extension

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

from __future__ import annotations

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["DictExt", "DictAnalyzeResult"]


# Library
# ---------------------------------------------------------------------------
import operator
from collections.abc import Callable
from typing import Any, Literal, NamedTuple, Self, TypeVar, overload

from absfuyu.core import GetClassMembersMixin, versionadded, versionchanged

# from absfuyu.typings import KT as _KT
# from absfuyu.typings import VT as _VT


# Type
# ---------------------------------------------------------------------------
KT = TypeVar("KT")
VT = TypeVar("VT")
VT2 = TypeVar("VT2")
R = TypeVar("R")  # Return type - Can be anything


# Class
# ---------------------------------------------------------------------------
class DictAnalyzeResult(NamedTuple):
    """
    Result for ``DictExt.analyze()``
    """

    max_value: int | float
    min_value: int | float
    max_list: list
    min_list: list


class DictExtLegacy(GetClassMembersMixin, dict[KT, VT]):
    """
    ``dict`` extension (with no generic type-hints)

    >>> # For a list of new methods
    >>> DictExt.show_all_methods()
    """

    @versionchanged("3.3.0", reason="Updated return type")
    def analyze(self) -> DictAnalyzeResult:
        """
        Analyze all the key values (``int``, ``float``)
        in ``dict`` then return highest/lowest index

        Returns
        -------
        dict
            Analyzed data


        Example:
        --------
        >>> test = DictExt({"abc": 9, "def": 9, "ghi": 8, "jkl": 1, "mno": 1})
        >>> test.analyze()
        DictAnalyzeResult(max_value=9, min_value=1, max_list=[('abc', 9), ('def', 9)], min_list=[('jkl', 1), ('mno', 1)])
        """
        try:
            dct: dict = self.copy()

            max_val: int | float = max(list(dct.values()))
            min_val: int | float = min(list(dct.values()))
            max_list = []
            min_list = []

            for k, v in dct.items():
                if v == max_val:
                    max_list.append((k, v))
                if v == min_val:
                    min_list.append((k, v))

            return DictAnalyzeResult(max_val, min_val, max_list, min_list)

        except TypeError:
            err_msg = "Value must be int or float"
            # logger.error(err_msg)
            raise ValueError(err_msg)  # noqa: B904

    def swap_items(self) -> Self:
        """
        Swap ``dict.keys()`` with ``dict.values()``

        Returns
        -------
        DictExt
            Swapped dict


        Example:
        --------
        >>> test = DictExt({"abc": 9})
        >>> test.swap_items()
        {9: 'abc'}
        """
        # return self.__class__(zip(self.values(), self.keys()))
        return self.__class__({v: k for k, v in self.items()})  # type: ignore

    def apply(self, func: Callable[[Any], Any], apply_to_value: bool = True) -> Self:
        """
        Apply function to ``DictExt.keys()`` or ``DictExt.values()``

        Parameters
        ----------
        func : Callable
            Callable function

        apply_to_value : bool
            | ``True``: Apply ``func`` to ``DictExt.values()``
            | ``False``: Apply ``func`` to ``DictExt.keys()``

        Returns
        -------
        DictExt
            DictExt


        Example:
        --------
        >>> test = DictExt({"abc": 9})
        >>> test.apply(str)
        {'abc': '9'}
        """
        if apply_to_value:
            new_dict = {k: func(v) for k, v in self.items()}
        else:
            new_dict = {func(k): v for k, v in self.items()}
        return self.__class__(new_dict)

    @versionchanged("5.0.0", reason="Updated to handle more types and operator")
    @versionadded("3.4.0")
    def aggregate(
        self,
        other_dict: dict[KT, VT],
        default_value: Any = 0,
        operator_func: Callable[[Any, Any], Any] = operator.add,  # operator add
    ) -> Self:
        """
        Aggregates the values of the current dictionary with another dictionary.

        For each unique key, this method applies the specified operator to the values
        from both dictionaries. If a key exists in only one dictionary, its value is used.
        If an error occurs during aggregation (e.g., incompatible types), the values
        from both dictionaries are returned as a list.

        Parameters
        ----------
        other_dict : dict
            The dictionary to aggregate with.

        default_value : Any, optional
            The value to use for missing keys, by default ``0``


        operator_func : Callable[[Any, Any], Any], optional
            A function that takes two arguments and returns a single value,
            by default ``operator.add``

        Returns
        -------
        Self
            A new instance of the aggregated dictionary.


        Example:
        --------
        >>> test = DictExt({"test": 5, "test2": 9})
        >>> agg = {"test1": 10, "test2": 1}
        >>> print(test.aggregate(agg))
        {'test1': 10, 'test': 5, 'test2': 10}

        >>> test = DictExt({"test": 5, "test2": 9})
        >>> agg = {"test1": 10, "test2": "1"}
        >>> print(test.aggregate(agg))
        {'test1': 10, 'test': 5, 'test2': [9, '1']}
        """
        merged_output = {}

        # Create a set of all unique keys from both dictionaries
        all_keys = set(self) | set(other_dict)

        for k in all_keys:
            # Retrieve values with default fallback
            value_self = self.get(k, default_value)
            value_other = other_dict.get(k, default_value)

            try:
                # Attempt to apply the operator for existing keys
                merged_output[k] = operator_func(value_self, value_other)
            except TypeError:
                # If a TypeError occurs (e.g., if values are not compatible), store values as a list
                merged_output[k] = [value_self, value_other]

        return self.__class__(merged_output)


# Type-hints are hard coded, no work around Self[KT, VT] yet
class DictExt(GetClassMembersMixin, dict[KT, VT]):
    """
    ``dict`` extension

    >>> # For a list of new methods
    >>> DictExt.show_all_methods()
    """

    # Analyze
    @versionchanged("3.3.0", reason="Updated return type")
    def analyze(self) -> DictAnalyzeResult:
        """
        Analyze all the key values (``int``, ``float``)
        in ``dict`` then return highest/lowest index

        Returns
        -------
        dict
            Analyzed data


        Example:
        --------
        >>> test = DictExt({"abc": 9, "def": 9, "ghi": 8, "jkl": 1, "mno": 1})
        >>> test.analyze()
        DictAnalyzeResult(max_value=9, min_value=1, max_list=[('abc', 9), ('def', 9)], min_list=[('jkl', 1), ('mno', 1)])
        """
        try:
            dct: dict = self.copy()

            max_val: int | float = max(list(dct.values()))
            min_val: int | float = min(list(dct.values()))
            max_list = []
            min_list = []

            for k, v in dct.items():
                if v == max_val:
                    max_list.append((k, v))
                if v == min_val:
                    min_list.append((k, v))

            return DictAnalyzeResult(max_val, min_val, max_list, min_list)

        except TypeError:
            err_msg = "Value must be int or float"
            # logger.error(err_msg)
            raise ValueError(err_msg)  # noqa: B904

    # Swap
    def swap_items(self) -> DictExt[VT, KT]:
        """
        Swap ``dict.keys()`` with ``dict.values()``

        Returns
        -------
        DictExt
            Swapped dict


        Example:
        --------
        >>> test = DictExt({"abc": 9})
        >>> test.swap_items()
        {9: 'abc'}
        """
        # return self.__class__(zip(self.values(), self.keys()))
        out: dict[VT, KT] = {v: k for k, v in self.items()}
        # return self.__class__({v: k for k, v in self.items()})  # type: ignore
        return self.__class__(out)  # type: ignore

    # Apply
    @overload
    def apply(self, func: Callable[[VT], R]) -> DictExt[KT, R]: ...

    @overload
    def apply(self, func: Callable[[KT], R], apply_to_value: Literal[False] = ...) -> DictExt[R, VT]: ...

    def apply(self, func: Callable[[KT | VT], R], apply_to_value: bool = True) -> DictExt[KT | R, VT | R]:  # type: ignore
        """
        Apply function to ``DictExt.keys()`` or ``DictExt.values()``

        Parameters
        ----------
        func : Callable
            Callable function

        apply_to_value : bool
            | ``True``: Apply ``func`` to ``DictExt.values()``
            | ``False``: Apply ``func`` to ``DictExt.keys()``

        Returns
        -------
        DictExt
            DictExt


        Example:
        --------
        >>> test = DictExt({"abc": 9})
        >>> test.apply(str)
        {'abc': '9'}
        """
        if apply_to_value:
            new_dict_v: dict[KT, R] = {k: func(v) for k, v in self.items()}
            return self.__class__(new_dict_v)  # type: ignore
        else:
            new_dict_k: dict[R, VT] = {func(k): v for k, v in self.items()}
            return self.__class__(new_dict_k)  # type: ignore

    # Aggregate
    @versionchanged("5.0.0", reason="Updated to handle more types and operator")
    @versionadded("3.4.0")
    def aggregate(
        self,
        other_dict: dict[KT, VT | VT2],
        default_value: Any = 0,
        operator_func: Callable[[Any, Any], R] = operator.add,  # operator add
    ):
        """
        Aggregates the values of the current dictionary with another dictionary.

        For each unique key, this method applies the specified operator to the values
        from both dictionaries. If a key exists in only one dictionary, its value is used.
        If an error occurs during aggregation (e.g., incompatible types), the values
        from both dictionaries are returned as a list.

        Parameters
        ----------
        other_dict : dict
            The dictionary to aggregate with.

        default_value : Any, optional
            The value to use for missing keys, by default ``0``


        operator_func : Callable[[Any, Any], Any], optional
            A function that takes two arguments and returns a single value,
            by default ``operator.add``

        Returns
        -------
        Self
            A new instance of the aggregated dictionary.


        Example:
        --------
        >>> test = DictExt({"test": 5, "test2": 9})
        >>> agg = {"test1": 10, "test2": 1}
        >>> print(test.aggregate(agg))
        {'test1': 10, 'test': 5, 'test2': 10}

        >>> test = DictExt({"test": 5, "test2": 9})
        >>> agg = {"test1": 10, "test2": "1"}
        >>> print(test.aggregate(agg))
        {'test1': 10, 'test': 5, 'test2': [9, '1']}
        """
        merged_output: dict[KT, VT | R | list[VT | VT2]] = {}

        # Create a set of all unique keys from both dictionaries
        all_keys = set(self) | set(other_dict)

        for k in all_keys:
            # Retrieve values with default fallback
            value_self = self.get(k, default_value)
            value_other = other_dict.get(k, default_value)

            try:
                # Attempt to apply the operator for existing keys
                merged_output[k] = operator_func(value_self, value_other)
            except TypeError:
                # If a TypeError occurs (e.g., if values are not compatible), store values as a list
                merged_output[k] = [value_self, value_other]  # type: ignore

        return self.__class__(merged_output)  # type: ignore
